/* md4-int.c - MD4 Message Digest Algorithm */

#include <stdio.h>

#include "general.h"
#include "osisec/alg_md4.h"

/* ARGSUSED */
int
md4_init_hash(hash, parms, key)
	struct MD4Hash **hash;
	struct MD4Parameters *parms;
	struct MD4Key *key;
{
	struct MD4Hash *result;

	result = (struct MD4Hash *) calloc(1, sizeof(*result));

	result->A = 0x67452301;
	result->B = 0xefcdab89;
	result->C = 0x98badcfe;
	result->D = 0x10325476;
	result->L = 0;

	*hash = result;

	return (OK);
}

#define F(x, y, z) (((x) & (y)) | ((~(x)) & (z)))
#define G(x, y, z) (((x) & (y)) | ((x) & (z)) | ((y) & (z)))
#define H(x, y, z) ((x)^(y)^(z))

#define ROT(x, s) ((((unsigned long)x) << (s)) | (((unsigned long)x) >> (32 - (s))))

#define GET_WORD(i)	(ubytes[4*i]+(ubytes[4*i+1]<<8)+(ubytes[4*i+2]<<16)+(ubytes[4*i+3]<<24))

#define R1(a, b, c, d, i, s) hash->a = \
ROT((hash->a + F(hash->b, hash->c, hash->d) + GET_WORD(i)), s)

#define R2(a, b, c, d, i, s) hash->a = \
ROT((hash->a + G(hash->b, hash->c, hash->d) + GET_WORD(i) + 0x5a827999), s)

#define R3(a, b, c, d, i, s) hash->a = \
ROT((hash->a + H(hash->b, hash->c, hash->d) + GET_WORD(i) + 0x6ed9eba1), s)

/* ARGSUSED */
int
md4_hash_block(hash, text, key, parms)
	struct MD4Hash *hash;
	char           *text;
	struct MD4Key *key;
	struct MD4Parameters *parms;
{
	unsigned long aa;
	unsigned long bb;
	unsigned long cc;
	unsigned long dd;
	unsigned char *ubytes = (unsigned char *) text;

	aa = hash->A;
	bb = hash->B;
	cc = hash->C;
	dd = hash->D;
	
	R1(A, B, C, D, 0, 3);
	R1(D, A, B, C, 1, 7);
	R1(C, D, A, B, 2, 11);
	R1(B, C, D, A, 3, 19);
	R1(A, B, C, D, 4, 3);
	R1(D, A, B, C, 5, 7);
	R1(C, D, A, B, 6, 11);
	R1(B, C, D, A, 7, 19);
	R1(A, B, C, D, 8, 3);
	R1(D, A, B, C, 9, 7);
	R1(C, D, A, B, 10, 11);
	R1(B, C, D, A, 11, 19);
	R1(A, B, C, D, 12, 3);
	R1(D, A, B, C, 13, 7);
	R1(C, D, A, B, 14, 11);
	R1(B, C, D, A, 15, 19);

	R2(A, B, C, D, 0, 3);
	R2(D, A, B, C, 4, 5);
	R2(C, D, A, B, 8, 9);
	R2(B, C, D, A, 12, 13);
	R2(A, B, C, D, 1, 3);
	R2(D, A, B, C, 5, 5);
	R2(C, D, A, B, 9, 9);
	R2(B, C, D, A, 13, 13);
	R2(A, B, C, D, 2, 3);
	R2(D, A, B, C, 6, 5);
	R2(C, D, A, B, 10, 9);
	R2(B, C, D, A, 14, 13);
	R2(A, B, C, D, 3, 3);
	R2(D, A, B, C, 7, 5);
	R2(C, D, A, B, 11, 9);
	R2(B, C, D, A, 15, 13);

	R3(A, B, C, D, 0, 3);
	R3(D, A, B, C, 8, 9);
	R3(C, D, A, B, 4, 11);
	R3(B, C, D, A, 12, 15);
	R3(A, B, C, D, 2, 3);
	R3(D, A, B, C, 10, 9);
	R3(C, D, A, B, 6, 11);
	R3(B, C, D, A, 14, 15);
	R3(A, B, C, D, 1, 3);
	R3(D, A, B, C, 9, 9);
	R3(C, D, A, B, 5, 11);
	R3(B, C, D, A, 13, 15);
	R3(A, B, C, D, 3, 3);
	R3(D, A, B, C, 11, 9);
	R3(C, D, A, B, 7, 11);
	R3(B, C, D, A, 15, 15);

	hash->A += aa;
	hash->B += bb;
	hash->C += cc;
	hash->D += dd;

	hash->L += 512;

	return (OK);
}

int
md4_display(hash)
	struct MD4Hash *hash;
{
	(void) printf("A = %04x\n", hash->A);
	(void) printf("B = %04x\n", hash->B);
	(void) printf("C = %04x\n", hash->C);
	(void) printf("D = %04x\n", hash->D);

	return (OK);
}


/* ARGSUSED */
int
md4_compare_hash(h1, h2, key, parms)
	struct MD4Hash *h1, *h2;
	struct MD4Key *key;
	struct MD4Parameters *parms;
{

	if (h1->A != h2->A)
		return (NOTOK);
	if (h1->B != h2->B)
		return (NOTOK);
	if (h1->C != h2->C)
		return (NOTOK);
	if (h1->D != h2->D)
		return (NOTOK);

	return (OK);
}

/* ARGSUSED */
int
md4_get_blocksize(parms)
	struct MD4Parameters *parms;
{
	return (512);
}

/* ARGSUSED */
int
md4_get_hashsize(parm)
	struct MD4Parameters *parm;
{
	return (128);
}


/* ARGSUSED */
int
md4_terminate_hash(hash, key, parms)
	struct MD4Hash *hash;
	struct MD4Key *key;
	struct MD4Parameters *parms;
{
	return (OK);
}

int
md4_hash_short_block(hash, buff, extra, key, parm)
	struct MD4Hash *hash;
	char           *buff;
	int             extra;
	struct MD4Key *key;
	struct MD4Parameters *parm;
{
	int             i;
	int             rc;

	buff[extra] = 0x80;
	for (i = extra+1; i < 64; i++)
		buff[i] = (char) 0;

	hash->L += extra*8;

	if (extra > 55) {
		(void) md4_hash_block(hash, buff, key, parm);
		bzero(buff, 64);
		hash->L -= 512; /* Fix up length as block was short */
	}
	buff[56] = hash->L & 0xff;
	buff[57] = (hash->L >>  8) & 0xff;
	buff[58] = (hash->L >> 16) & 0xff;
	buff[59] = (hash->L >> 24) & 0xff;
	buff[60] = 0;
	buff[61] = 0;
	buff[62] = 0;
	buff[63] = 0;

	rc = md4_hash_block(hash, buff, key, parm);
	return (rc);
}


/* ARGSUSED */
int
md4_pack_hash(hash, out, n_bits, parm)
	struct MD4Hash *hash;
	char          **out;
	int            *n_bits;
	struct MD4Parameters *parm;
{
	char           *cp;

	cp = calloc(1, 16);
	if (cp == NULLCP)
		return (NOTOK);

	cp[0] = hash->A & 255;
	cp[1] = (hash->A >> 8) & 255;
	cp[2] = (hash->A >> 16) & 255;
	cp[3] = (hash->A >> 24) & 255;

	cp[4] = hash->B & 255;
	cp[5] = (hash->B >> 8) & 255;
	cp[6] = (hash->B >> 16) & 255;
	cp[7] = (hash->B >> 24) & 255;

	cp[8] = hash->C & 255;
	cp[9] = (hash->C >> 8) & 255;
	cp[10] = (hash->C >> 16) & 255;
	cp[11] = (hash->C >> 24) & 255;

	cp[12] = hash->D & 255;
	cp[13] = (hash->D >> 8) & 255;
	cp[14] = (hash->D >> 16) & 255;
	cp[15] = (hash->D >> 24) & 255;

	(*n_bits) = 128;
	(*out) = cp;

	return (OK);
}


int
md4_unpack_hash(cp, hash)
	unsigned char  *cp;
	struct MD4Hash **hash;
{
	struct MD4Hash *result;

	result = (struct MD4Hash *) calloc(1, sizeof(*result));
	if (result == (struct MD4Hash *) 0)
		return (NOTOK);

	result->A = cp[0] | (cp[1] << 8) | (cp[2] << 16) | (cp[3] << 24);
	result->B = cp[4] | (cp[5] << 8) | (cp[6] << 16) | (cp[7] << 24);
	result->C = cp[8] | (cp[9] << 8) | (cp[10] << 16) | (cp[11] << 24);
	result->D = cp[12] | (cp[13] << 8) | (cp[14] << 16) | (cp[15] << 24);

	(*hash) = result;

	return (OK);
}

/* ARGSUSED */
int
md4_free_hash(hash, parm)
	struct MD4Hash *hash;
	struct MD4Parameters *parm;
{
	free((char *) hash);
	return (OK);
}

