/* md5-int.c - MD5 Message Digest Algorithm */

#include <stdio.h>

#include "general.h"
#include "osisec/alg_md5.h"

/* ARGSUSED */
int
md5_init_hash(hash, parms, key)
	struct MD5Hash **hash;
	struct MD5Parameters *parms;
	struct MD5Key *key;
{
	struct MD5Hash *result;

	result = (struct MD5Hash *) calloc(1, sizeof(*result));

	result->A = 0x67452301;
	result->B = 0xefcdab89;
	result->C = 0x98badcfe;
	result->D = 0x10325476;
	result->L = 0;

	*hash = result;

	return (OK);
}

#define F(x, y, z) (((x) & (y)) | ((~(x)) & (z)))
#define G(x, y, z) (((x) & (z)) | ((y) & (~(z))))
#define H(x, y, z) ((x)^(y)^(z))
#define I(x, y, z) ((y)^((x)|(~(z))))

#define ROT(x, s) ((((unsigned long)x) << (s)) | (((unsigned long)x) >> (32 - (s))))

#define GET_WORD(i)	(ubytes[4*i]+(ubytes[4*i+1]<<8)+(ubytes[4*i+2]<<16)+(ubytes[4*i+3]<<24))

#define R1(a, b, c, d, i, s, t) hash->a = \
hash->b + ROT((hash->a + F(hash->b, hash->c, hash->d) + GET_WORD(i) + t), s)

#define R2(a, b, c, d, i, s, t) hash->a = \
hash->b + ROT((hash->a + G(hash->b, hash->c, hash->d) + GET_WORD(i) + t), s)

#define R3(a, b, c, d, i, s, t) hash->a = \
hash->b + ROT((hash->a + H(hash->b, hash->c, hash->d) + GET_WORD(i) + t), s)

#define R4(a, b, c, d, i, s, t) hash->a = \
hash->b + ROT((hash->a + I(hash->b, hash->c, hash->d) + GET_WORD(i) + t), s)

#define S11	7
#define S12	12
#define S13	17
#define S14	22

#define S21	5
#define	S22	9
#define S23	14
#define S24	20

#define S31	4
#define S32	11
#define S33	16
#define S34	23

#define S41	6
#define S42	10
#define S43	15
#define S44	21

/* ARGSUSED */
int
md5_hash_block(hash, text, key, parms)
	struct MD5Hash *hash;
	char           *text;
	struct MD5Key *key;
	struct MD5Parameters *parms;
{
	unsigned long aa;
	unsigned long bb;
	unsigned long cc;
	unsigned long dd;
	unsigned char *ubytes = (unsigned char *) text;

	aa = hash->A;
	bb = hash->B;
	cc = hash->C;
	dd = hash->D;
	
	R1(A, B, C, D,  0, S11, 3614090360);
	R1(D, A, B, C,  1, S12, 3905402710);
	R1(C, D, A, B,  2, S13,  606105819);
	R1(B, C, D, A,  3, S14, 3250441966);
	R1(A, B, C, D,  4, S11, 4118548399);
	R1(D, A, B, C,  5, S12, 1200080426);
	R1(C, D, A, B,  6, S13, 2821735955);
	R1(B, C, D, A,  7, S14, 4249261313);
	R1(A, B, C, D,  8, S11, 1770035416);
	R1(D, A, B, C,  9, S12, 2336552879);
	R1(C, D, A, B, 10, S13, 4294925233);
	R1(B, C, D, A, 11, S14, 2304563134);
	R1(A, B, C, D, 12, S11, 1804603682);
	R1(D, A, B, C, 13, S12, 4254626195);
	R1(C, D, A, B, 14, S13, 2792965006);
	R1(B, C, D, A, 15, S14, 1236535329);

	R2(A, B, C, D,  1, S21, 4129170786);
	R2(D, A, B, C,  6, S22, 3225465664);
	R2(C, D, A, B, 11, S23,  643717713);
	R2(B, C, D, A,  0, S24, 3921069994);
	R2(A, B, C, D,  5, S21, 3593408605);
	R2(D, A, B, C, 10, S22,   38016083);
	R2(C, D, A, B, 15, S23, 3634488961);
	R2(B, C, D, A,  4, S24, 3889429448);
	R2(A, B, C, D,  9, S21,  568446438);
	R2(D, A, B, C, 14, S22, 3275163606);
	R2(C, D, A, B,  3, S23, 4107603335);
	R2(B, C, D, A,  8, S24, 1163531501);
	R2(A, B, C, D, 13, S21, 2850285829);
	R2(D, A, B, C,  2, S22, 4243563512);
	R2(C, D, A, B,  7, S23, 1735328473);
	R2(B, C, D, A, 12, S24, 2368359562);

	R3(A, B, C, D,  5, S31, 4294588738);
	R3(D, A, B, C,  8, S32, 2272392833);
	R3(C, D, A, B, 11, S33, 1839030562);
	R3(B, C, D, A, 14, S34, 4259657740);
	R3(A, B, C, D,  1, S31, 2763975236);
	R3(D, A, B, C,  4, S32, 1272893353);
	R3(C, D, A, B,  7, S33, 4139469664);
	R3(B, C, D, A, 10, S34, 3200236656);
	R3(A, B, C, D, 13, S31,  681279174);
	R3(D, A, B, C,  0, S32, 3936430074);
	R3(C, D, A, B,  3, S33, 3572445317);
	R3(B, C, D, A,  6, S34,   76029189);
	R3(A, B, C, D,  9, S31, 3654602809);
	R3(D, A, B, C, 12, S32, 3873151461);
	R3(C, D, A, B, 15, S33,  530742520);
	R3(B, C, D, A,  2, S34, 3299628645);

	R4(A, B, C, D,  0, S41, 4096336452);
	R4(D, A, B, C,  7, S42, 1126891415);
	R4(C, D, A, B, 14, S43, 2878612391);
	R4(B, C, D, A,  5, S44, 4237533241);
	R4(A, B, C, D, 12, S41, 1700485571);
	R4(D, A, B, C,  3, S42, 2399980690);
	R4(C, D, A, B, 10, S43, 4293915773);
	R4(B, C, D, A,  1, S44, 2240044497);
	R4(A, B, C, D,  8, S41, 1873313359);
	R4(D, A, B, C, 15, S42, 4264355552);
	R4(C, D, A, B,  6, S43, 2734768916);
	R4(B, C, D, A, 13, S44, 1309151649);
	R4(A, B, C, D,  4, S41, 4149444226);
	R4(D, A, B, C, 11, S42, 3173756917);
	R4(C, D, A, B,  2, S43,  718787259);
	R4(B, C, D, A,  9, S44, 3951481745);

	hash->A += aa;
	hash->B += bb;
	hash->C += cc;
	hash->D += dd;

	hash->L += 512;

	return (OK);
}

int
md5_display(hash)
	struct MD5Hash *hash;
{
	(void) printf("A = %04x\n", hash->A);
	(void) printf("B = %04x\n", hash->B);
	(void) printf("C = %04x\n", hash->C);
	(void) printf("D = %04x\n", hash->D);

	return (OK);
}


/* ARGSUSED */
int
md5_compare_hash(h1, h2, key, parms)
	struct MD5Hash *h1, *h2;
	struct MD5Key *key;
	struct MD5Parameters *parms;
{

	if (h1->A != h2->A)
		return (NOTOK);
	if (h1->B != h2->B)
		return (NOTOK);
	if (h1->C != h2->C)
		return (NOTOK);
	if (h1->D != h2->D)
		return (NOTOK);

	return (OK);
}

/* ARGSUSED */
int
md5_get_blocksize(parms)
	struct MD5Parameters *parms;
{
	return (512);
}

/* ARGSUSED */
int
md5_get_hashsize(parm)
	struct MD5Parameters *parm;
{
	return (128);
}


/* ARGSUSED */
int
md5_terminate_hash(hash, key, parms)
	struct MD5Hash *hash;
	struct MD5Key *key;
	struct MD5Parameters *parms;
{
	return (OK);
}

int
md5_hash_short_block(hash, buff, extra, key, parm)
	struct MD5Hash *hash;
	char           *buff;
	int             extra;
	struct MD5Key *key;
	struct MD5Parameters *parm;
{
	int             i;
	int             rc;

	buff[extra] = 0x80;
	for (i = extra+1; i < 64; i++)
		buff[i] = (char) 0;

	hash->L += extra*8;

	if (extra > 55) {
		(void) md5_hash_block(hash, buff, key, parm);
		bzero(buff, 64);
		hash->L -= 512; /* Fix up length as block was short */
	}
	buff[56] = hash->L & 0xff;
	buff[57] = (hash->L >>  8) & 0xff;
	buff[58] = (hash->L >> 16) & 0xff;
	buff[59] = (hash->L >> 24) & 0xff;
	buff[60] = 0;
	buff[61] = 0;
	buff[62] = 0;
	buff[63] = 0;

	rc = md5_hash_block(hash, buff, key, parm);
	return (rc);
}


/* ARGSUSED */
int
md5_pack_hash(hash, out, n_bits, parm)
	struct MD5Hash *hash;
	char          **out;
	int            *n_bits;
	struct MD5Parameters *parm;
{
	char           *cp;

	cp = calloc(1, 16);
	if (cp == NULLCP)
		return (NOTOK);

	cp[0] = hash->A & 255;
	cp[1] = (hash->A >> 8) & 255;
	cp[2] = (hash->A >> 16) & 255;
	cp[3] = (hash->A >> 24) & 255;

	cp[4] = hash->B & 255;
	cp[5] = (hash->B >> 8) & 255;
	cp[6] = (hash->B >> 16) & 255;
	cp[7] = (hash->B >> 24) & 255;

	cp[8] = hash->C & 255;
	cp[9] = (hash->C >> 8) & 255;
	cp[10] = (hash->C >> 16) & 255;
	cp[11] = (hash->C >> 24) & 255;

	cp[12] = hash->D & 255;
	cp[13] = (hash->D >> 8) & 255;
	cp[14] = (hash->D >> 16) & 255;
	cp[15] = (hash->D >> 24) & 255;

	(*n_bits) = 128;
	(*out) = cp;

	return (OK);
}


int
md5_unpack_hash(cp, hash)
	unsigned char  *cp;
	struct MD5Hash **hash;
{
	struct MD5Hash *result;

	result = (struct MD5Hash *) calloc(1, sizeof(*result));
	if (result == (struct MD5Hash *) 0)
		return (NOTOK);

	result->A = cp[0] | (cp[1] << 8) | (cp[2] << 16) | (cp[3] << 24);
	result->B = cp[4] | (cp[5] << 8) | (cp[6] << 16) | (cp[7] << 24);
	result->C = cp[8] | (cp[9] << 8) | (cp[10] << 16) | (cp[11] << 24);
	result->D = cp[12] | (cp[13] << 8) | (cp[14] << 16) | (cp[15] << 24);

	(*hash) = result;

	return (OK);
}

/* ARGSUSED */
int
md5_free_hash(hash, parm)
	struct MD5Hash *hash;
	struct MD5Parameters *parm;
{
	free((char *) hash);
	return (OK);
}

