/* alg_rsa.h - Access Header for OSISEC-conformant implementation of RSA */

#ifndef _OSISEC_RSA_DEFINED_
#define _OSISEC_RSA_DEFINED_

#include "osisec/osisec.h"
#include "osisec/keysize.h"

typedef void *RSAEncrypted;
#define NULLENCRYPTED ((unsigned*)0)

struct RSASecretKey {
	unsigned       *p;
	unsigned       *q;
	unsigned       *a;
	unsigned       *n;
	unsigned       *d;
	unsigned       *d1;
	unsigned       *d2;
};

struct RSAPublicKey {
	unsigned       *n;
	unsigned        e;
};

struct RSAParameters {
	int             len;
	int             shift;
};

extern int      rsa_encrypt();
extern int      rsa_decrypt();
extern int      rsa_read_secret_key();
extern int      rsa_write_secret_key();
extern int      rsa_read_public_key();
extern int      rsa_write_public_key();
extern int      rsa_pack_encrypted();
extern int      rsa_unpack_encrypted();
extern int      rsa_pack_public_key();
extern int      rsa_unpack_public_key();
extern int      rsa_pack_secret_key();
extern int      rsa_unpack_secret_key();
extern int      rsa_get_blocksize();
extern int      rsa_free_encrypted();

#define RSAPUBLICKEYALGORITHMVALUE { \
	"rsa", \
	rsa_encrypt, \
	rsa_decrypt, \
	rsa_read_secret_key, \
	rsa_write_secret_key, \
	rsa_read_public_key, \
	rsa_write_public_key, \
	rsa_pack_encrypted, \
	rsa_unpack_encrypted, \
	rsa_pack_public_key, \
	rsa_unpack_public_key, \
	rsa_pack_secret_key, \
	rsa_unpack_secret_key, \
	rsa_get_blocksize, \
	rsa_free_encrypted \
}

extern struct PublicKeyAlgorithm	*use_pk_rsa();
extern struct EncryptionAlgorithm	*use_enc_rsa();

extern struct EncryptionAlgorithm rsa_enc_alg;

#ifndef _MD2WITHRSA_
#define _MD2WITHRSA_
extern struct SignatureAlgorithm md2withrsa_sig_alg;
extern struct SignatureAlgorithm* use_sig_md2withrsa();
#endif

#ifndef _MD4WITHRSA_
#define _MD4WITHRSA_
extern struct SignatureAlgorithm md4withrsa_sig_alg;
extern struct SignatureAlgorithm* use_sig_md4withrsa();
#endif

#ifndef _MD4AWITHRSA_
#define _MD4AWITHRSA_
extern struct SignatureAlgorithm md4awithrsa_sig_alg;
extern struct SignatureAlgorithm* use_sig_md4awithrsa();
#endif

#ifndef _MD5WITHRSA_
#define _MD5WITHRSA_
extern struct SignatureAlgorithm md5withrsa_sig_alg;
extern struct SignatureAlgorithm* use_sig_md5withrsa();
#endif

#endif
