/* chr.c - Exponentiation modulo pq using Chinese Remainder Theorem */

#include "rsa.h"

void 
chexp(r, x, n, p, p1, pshift, q, q1, qshift, d1, d2, a, ll, l, b1, b2, u, v, work, table)
	unsigned        r[];	/* The result      */
	unsigned        x[];	/* Ciphertext      */
	unsigned        n[];	/* The modulus     */
	unsigned        p[];	/* Factor of n     */
	unsigned        p1[];	/* p shifted left  */
	int             pshift;	/* amount shifted  */
	unsigned        q[];	/* Factor of  n    */
	unsigned        q1[];	/* q shifted left  */
	int             qshift;	/* amount shifted  */
	unsigned        d1[];	/* Power mod (p-1) */
	unsigned        d2[];	/* Power mod (q-1) */
	unsigned        a[];	/* p^-1 mod q      */
	int             ll;	/* length of n     */
	int             l;	/* length of p,q   */
	unsigned        b1[];
	unsigned        b2[];
	unsigned        u[];	/* work space      */
	unsigned        v[];
	unsigned        work[];
	unsigned       *table[];
{
	int             i;
	int             cy;

	bsl(p1, p, l, pshift);
	bsl(work, x, ll, pshift);
	fullreduce(work, ll - l, p1, l);
	fastexp(b1, work, d1, p1, l, pshift, u, v, table);

	bsl(q1, q, l, qshift);
	bsl(work, x, ll, qshift);
	fullreduce(work, ll - l, q1, l);
	fastexp(b2, work, d2, q1, l, qshift, u, v, table);

	bsr(r, b1, l, pshift);
	for (i = l + 1; i <= ll; i++)
		r[i] = 0;
	bsl(u, r, l, qshift);

	cy = sub(u, b2, l);
	if (cy != 0)
		(void) add(u, q1, l);

	copy(work, u, l);
	modmul(u, work, a, q1, l);

	multiply(work, u, p, l);
	bsr(work, work, ll, qshift);
	cy = sub(r, work, ll);
	if (cy != 0)
		(void)add(r, n, ll);
}
