/* testmath.c - Validate extended precision modular arithmetic routines */

#include <stdio.h>
#include "osisec/numalgs.h"
#include "testsize.h"

void
starting(test)
	char           *test;
{
	(void) printf("Testing `%s'\n", test);
}

void
failed(test, reason)
	char           *test;
	char           *reason;
{
	void            exit();

	(void) printf("Test of `%s' failed - %s\n", test, reason);
	exit(1);
}

int
main()
{
	unsigned        a[TESTSIZE];
	unsigned        b[TESTSIZE];
	unsigned        c[TESTSIZE];
	unsigned        d[TESTSIZE];
	unsigned        w1[TESTSIZE];
	unsigned        w2[TESTSIZE];

	unsigned        modulus[TESTSIZE];
	int             i;
	int             shift;
	void            exit();

	(void) printf("Testing with %d bit integers.\n", TESTSIZE * BITS);

	starting("add");
	clear(a, TESTSIZE - 1);
	clear(b, TESTSIZE - 1);
	a[LSW] = MASK;
	b[LSW] = 1;
	(void) add(a, b, TESTSIZE - 1);
	if (a[LSW] != 0)
		failed("add", "increment");
	if (a[LSW + 1] != 1)
		failed("add", "carry");
	if (a[LSW + 2] != 0)
		failed("add", "carry propogation");

	starting("bsr");
	bsr(b, a, TESTSIZE - 1, 1);
	if (b[LSW] != TOPB)
		failed("bsr", "1");
	if (b[LSW + 1] != 0)
		failed("bsr", "2");

	starting("bsl");
	bsl(a, b, TESTSIZE - 1, 1);
	if (a[LSW] != 0)
		failed("bsl", "1");
	if (a[LSW + 1] != 1)
		failed("bsl", "2");

	starting("reduce");
	shift = 1;
	for (i = LSW; i <= TESTSIZE - 1; i++)
		modulus[i] = MASK;
	modulus[LSW] = (MASK << shift) & MASK;
	clear(a, TESTSIZE - 1);
	a[LSW] = 2 << shift;
	reduce(a, 5, modulus, TESTSIZE - 1);
	if (a[LSW] != 7 << shift)
		failed("reduce", "1");

	starting("modmul");
	copy(a, modulus, TESTSIZE - 1);
	(void) sbs(a, TESTSIZE - 1, (unsigned)(1 << shift));
	clear(b, TESTSIZE - 1);
	b[LSW] = 5;
	modmul(c, a, b, modulus, TESTSIZE - 1);
	(void) ads(c, TESTSIZE - 1, (unsigned)(5 << shift));
	for (i =
	     LSW; i <= TESTSIZE - 1; i++)
		if (modulus[i] != c[i])
			failed("modmul", "1");

	starting("modexp");
	(void) printf("Note - 2^%d-1 must be a prime for this test to work.\n",
		      BITS * TESTSIZE - shift);
	clear(a, TESTSIZE - 1);
	a[LSW] = 5
		<< shift;
	bsr(b, modulus, TESTSIZE - 1, shift);
	modexp(c, a, b,
	       modulus, TESTSIZE - 1, shift, w1, w2);
	if (c[LSW] != 5 << shift)
		failed("modexp", "1");
	if (c[LSW + 1] != 0)
		failed("modexp", "2");

	starting("smallexp");
	clear(a, TESTSIZE - 1);
	a[LSW] = 5 << shift;
	clear(b, TESTSIZE - 1);
	b[LSW] = 2;
	modexp(c, a, b, modulus,
	       TESTSIZE - 1, shift, w1, w2);
	clear(a, TESTSIZE - 1);
	a[LSW] = 5 <<
		shift;
	clear(b, TESTSIZE - 1);
	b[LSW] = 2;
	smallexp(d, a, 2, modulus,
		 TESTSIZE - 1, shift, w1, w2);
	hexdump(c, TESTSIZE - 1);
	hexdump(d, TESTSIZE - 1);
	for (i = 0; i < TESTSIZE; i++)
		if (c[i] != d[i])
			failed("smallexp", "1");
	(void) printf("All tests run OK.\n");
	exit(0);
	return 0;
}
