/* testrsa.c - validate the RSA encryption/decryption routines */

#include <stdio.h>
#include <sys/types.h>
#include <sys/times.h>

#include "general.h"
#include "osisec/numalgs.h"
#include "testsize.h"

gulp(f)
	FILE           *f;
{
	int             i;
	for (i = 0; i < 5; i++)
		(void) getc(f);
}

hexin(f, a, l)
	FILE           *f;
	unsigned        a[];
	int             l;
{
	int             i;
	int             j;
	int             n;
	char            c;

	for (i = l; i >= 0; i--) {
		a[i] = 0;
		for (j = 0; j < NPW; j++) {
			c = getc(f);
			n = (int) c - (int) '0';
			if (c > '9')
				n = (int) c - (int) 'a' + 10;
			a[i] = (a[i] << 4) + n;
		}
	}
	(void)fscanf(f, " \n");
}

int
main(argc, argv)
	int             argc;
	char          **argv;
{
	int             i;
	FILE           *f1;
	unsigned        p[HALFKEYSIZE];
	unsigned        p1[HALFKEYSIZE];
	unsigned        q[HALFKEYSIZE];
	unsigned        q1[HALFKEYSIZE];
	unsigned        a[HALFKEYSIZE];
	unsigned        n[KEYSIZE];
	unsigned        d[KEYSIZE];
	unsigned        x[HALFKEYSIZE];
	unsigned        y[HALFKEYSIZE];
	unsigned        m[KEYSIZE];
	unsigned        u[KEYSIZE];
	unsigned        v[KEYSIZE];
	unsigned        b[KEYSIZE];
	unsigned        c[KEYSIZE];
	unsigned        b1[KEYSIZE];
	unsigned        b2[KEYSIZE];
	unsigned        t1[HALFKEYSIZE];
	unsigned        r[KEYSIZE];
	unsigned       *table[HALFKEYSIZE];
	int             shift;
	int             pshift;
	int             qshift;
	struct tms      tbuf;
	time_t          now;
	float           secs;
	void		exit();
	int		dummy;	

	if(argc != 2)
		f1 = fopen("rsakey", "r");
	else
		f1 = fopen(argv[1], "r");
	if (f1 == NULL) {
		(void)printf("Can't open rsakey file");
		exit(1);
	}
	(void)fscanf(f1, "L  = %d\n", &dummy);
	(void)printf("P = ");
	gulp(f1);
	hexin(f1, p, HALFKEYSIZE - 1);
	hexdump(p, HALFKEYSIZE - 1);
	(void)printf("Q = ");
	gulp(f1);
	hexin(f1, q, HALFKEYSIZE - 1);
	hexdump(q, HALFKEYSIZE - 1);
	(void)printf("A = ");
	gulp(f1);
	hexin(f1, a, HALFKEYSIZE - 1);
	hexdump(a, HALFKEYSIZE - 1);
	(void)printf("N = ");
	gulp(f1);
	hexin(f1, n, KEYSIZE - 1);
	hexdump(n, KEYSIZE - 1);
	(void)printf("D = ");
	gulp(f1);
	hexin(f1, d, KEYSIZE - 1);
	hexdump(d, KEYSIZE - 1);
	(void)printf("D1= ");
	gulp(f1);
	hexin(f1, x, HALFKEYSIZE - 1);
	hexdump(x, HALFKEYSIZE - 1);
	(void)printf("D2= ");
	gulp(f1);
	hexin(f1, y, HALFKEYSIZE - 1);
	hexdump(y, HALFKEYSIZE - 1);

	for (i = 0; i <= HALFKEYSIZE - 1; i++)
		table[i] = (unsigned *) calloc((unsigned)(4 * (HALFKEYSIZE - 1 + 1)), sizeof(unsigned));

	shift = bitsfree(n, KEYSIZE - 1);
	bsl(m, n, KEYSIZE - 1, shift);

	pshift = bitsfree(p, HALFKEYSIZE - 1);
	bsl(p1, p, HALFKEYSIZE - 1, pshift);
	qshift = bitsfree(q, HALFKEYSIZE - 1);
	bsl(q1, q, HALFKEYSIZE - 1, qshift);

	clear(b, KEYSIZE - 1);
	(void)srand((int)time((time_t*)0));
	large_random(b, KEYSIZE - 1);
	b[KEYSIZE - 1] = n[KEYSIZE - 1] - 1;
	copy(c, b, KEYSIZE - 1);
	(void)printf("Plaintext = ");
	hexdump(b, KEYSIZE - 1);
	bsl(b, b, KEYSIZE - 1, shift);

	(void)times(&tbuf);
	now = tbuf.tms_utime;

#ifdef FOOBAR
	fermatexp(r, b, 15, m, KEYSIZE - 1, shift, u, v);
#else
	smallexp(r, b, 65537, m, KEYSIZE - 1, shift, u, v);
#endif
	bsr(b, r, KEYSIZE - 1, shift);

	(void)times(&tbuf);
	now = tbuf.tms_utime - now;
	secs = ((float) now) / HZ;
	(void)printf("Encryption time = %.2f s.\n", secs);

	(void)times(&tbuf);
	now = tbuf.tms_utime;

	(void)printf("Ciphertext = ");
	hexdump(b, KEYSIZE - 1);

	bsl(b1, b, KEYSIZE - 1, pshift);
	bsl(b2, b, KEYSIZE - 1, qshift);

	fullreduce(b1, KEYSIZE - HALFKEYSIZE, p1, HALFKEYSIZE - 1);
	fullreduce(b2, KEYSIZE - HALFKEYSIZE, q1, HALFKEYSIZE - 1);

	/* Decrypt using library routine. */

	(void)times(&tbuf);
	now = (tbuf.tms_utime);

	chexp(r, b, n, p, p1, pshift, q, q1, qshift, x, y, a, KEYSIZE - 1, HALFKEYSIZE - 1, b1, b2, u, v, t1, table);

	(void)times(&tbuf);
	now = tbuf.tms_utime - now;
	secs = ((float) now) / HZ;
	(void)printf("Decryption time = %.2f s.\n", secs);

	(void)printf("Result = ");
	hexdump(r, KEYSIZE - 1);

	if (cpdelta(r, c, KEYSIZE - 1, 0) != 0) {
		(void)printf("Comparsion error!\n");
		exit(1);
		return 1;
	}

	(void) malloc(1);	/* check heap consistency */
	exit(0);
	return 0;
}
