/*
 * $Id: bp.c,v 1.11 1994/05/21 13:36:38 katsu Exp $
 *
 * Copyright (c) 1993 Katsuaki Suzuki <katsu@wide.ad.jp>
 *                    Takashi Taniguchi <tani@wide.ad.jp>
 *     Thanks to WIDE Project WPNC Working Group
 *               The University of Electro Communications 
 *               All rights reserved.
 */

#undef DEBUG

#include <stdio.h>
#include <signal.h>
#include <setjmp.h>
#include <ctype.h>

#ifndef luna
#include <unistd.h>
#endif

#ifdef FILIO_H
# include <sys/filio.h>
#else
# include <sys/ioctl.h>
#endif

#if defined(USE_TERMIO)
# if defined(SYSV_TERMIO)
#  include <termio.h>
# else
#  include <termios.h>
# endif
#else				/* traditional V7 ioctl */
# include <sgtty.h>
#endif

#include <errno.h>
#include <fcntl.h>

#include <sys/param.h>
#include <time.h>
#include "bplus.h"

char title[] = "Bplus client Version 1.20 <katsu@wide.ad.jp>";

FILE	*tty, *ttyr;

int infd = 0, outfd = 1;
int interactive = 0;
int userabort;
char s128[128];
char esci_response[] = "#XXX,PB";

extern int PortBps;
extern int BP_Auto_Resume;
extern int BP_Use_File_Size;
extern long BP_File_Size;
int terminated;

#if defined(USE_TERMIO)
static struct termios std, buf, buf2;
#else
static struct sgttyb std, buf, buf2;
#endif

int displen = 0;

struct _STVALUE {
    char *fmt;
    long value;
    int flags;
} stval[STNumVal] = {
    {"Message:   %s\n",0,1},
    {"Direction: %s\n",0,1},
    {"Type:      %s\n",0,1},
    {"File:      %s\n",0,1},

    {"CS(%ld) ",0,4},
    {"PS(%ld) ",0,4},
    {"ES(%ld) ",0,1},
    {"CR(%ld) ",0,4},
    {"PR(%ld) ",0,4},
    {"ER(%ld) ",0,1},
    {"RS(%8ld) ",0,4},
    {"Size(%8ld) ",0,4},
    {"Data(%8ld) ",0,1},
    {"CR(%ld) ",0,4},
    {"Rate(%4ld) ",0,0},
    {"Elap(%6ld) ",0,0},
    {"RTime(%5ld) ",0,0}
};


/*
 * raw2cock()
 * restore original terminal mode.
 */
static void
raw2cock()
{
#if defined(USE_TERMIO)
    tcsetattr(infd, TCSANOW, &buf);
    tcsetattr(outfd, TCSANOW, &buf2);
#else
    ioctl(infd, TIOCSETP, &buf);
    ioctl(outfd, TIOCSETP, &buf2);
#endif

}

/*
 * cock2raw()
 * change original terminal mode to raw mode.
 */
#if defined(USE_TERMIO)
static void
rawmode(t)
struct termios *t;
{
    t->c_cc[VTIME] = 1;
    t->c_cc[VMIN] = 0;
    t->c_iflag &= ~(INLCR|ICRNL|ISTRIP|IXON|BRKINT|IGNCR|IXOFF);
    t->c_oflag &= ~OPOST;
    t->c_lflag &= ~(ICANON|ISIG|ECHO|ECHOE|ECHOK|ECHONL|IEXTEN);
    t->c_cflag &= ~(CSIZE|PARENB);
    t->c_cflag |= CS8;
}
#endif

static void
cock2raw()
{
#if defined(USE_TERMIO)
    tcgetattr(infd, &buf);
    tcgetattr(outfd, &buf2);

    std = buf;
    rawmode(&std);
    tcsetattr(infd, TCSANOW, &std);

    std = buf2;
    rawmode(&std);
    tcsetattr(outfd, TCSANOW, &std);
#else
    ioctl(infd, TIOCGETP, &buf);
    ioctl(outfd, TIOCGETP, &buf2);

    std = buf;
    std.sg_flags = RAW;
    ioctl(infd, TIOCSETP, &std);
  
    std = buf2;
    std.sg_flags = RAW;
    ioctl(outfd, TIOCSETP, &std);
#endif
}

void
cancelsig(sig)
int sig;
{
    userabort = 1;
    fprintf(tty, "bp:Interrupted!!\n");
    alarm(0);
    alarm(10);
}

void
abortsig(sig)
int sig;
{
    fprintf(tty, "bp:ABORT!!\n");
    raw2cock();
    exit(128 + sig);
}

void
toabortsig(sig)
int sig;
{
    fprintf(tty, "bp:TIME OUT ABORT!!\n");
    raw2cock();
    exit(128 + sig);
}

static void
setsignals()
{
    signal(SIGALRM, toabortsig);
    signal(SIGINT, cancelsig);
    signal(SIGTERM, cancelsig);
    signal(SIGQUIT, cancelsig);
}

static void
dflsignals()
{
    signal(SIGINT, abortsig);
    signal(SIGTERM, abortsig);
    signal(SIGQUIT, abortsig);
}


/* send a char asynchronous */
Async_Send(ch)
unsigned char ch;
{
    ch &= 0xff;

#ifdef DEBUG
    fprintf(tty, "*\r");
#endif
    while(write(outfd, &ch, 1) != 1 );
}

/* check and read char
 * return 1:	ch <- char
 * return 0:	ch undefined
 */

Async_BuffeR_Check(ch)
unsigned char *ch;
{
    static int nrest = 0;
    static char rbuff[65536], *rbp;

    if (nrest > 0){
	--nrest;
	*ch = *rbp++;
#ifdef DEBUG
	fprintf(tty, "(%02x ", *ch);
#endif
	return 1;
    }
#if !(defined(USE_TERMIO) || defined(SYSV_TERMIO)) || defined(__386BSD__)
    ioctl(infd, FIONREAD, &nrest);
    if (nrest <= 0)
	return 0;
#endif
    nrest = read(infd, rbuff, sizeof rbuff);
    if (nrest <= 0)
	return 0;
    --nrest;
    rbp = rbuff;
    *ch = *rbp++;
#ifdef DEBUG
    fprintf(tty, "<%02x ", *ch);
#endif
    return 1;
}

/* check user abort
 * return !0 if abort
 */
ST_Check_Abort()
{
    int rval;

    rval = userabort;
    userabort = 0;
    return rval;
}

ST_Display_String(type, mes)
int type;
char *mes;
{
    if (displen) {
	fputc('\n', tty);
	displen = 0;
    }
    fprintf(tty, stval[type].fmt, mes);
}

ST_Display_Value(type, val)
int type;
long val;
{
    register struct _STVALUE *sp = &stval[type];

    sp->value = val;
    if (!(sp->flags & 1))
	return;
    if (sp->flags & 2) {
	if (displen) {
	    fputc('\n', tty);
	    displen = 0;
	}
	fprintf(tty, sp->fmt, sp->value);
    } else {
	int i;
	char buff[BUFSIZ], *p;

	p = buff;
	*p = '\0';
	sprintf( p, "Filesize:%d ", stval[10].value+stval[11].value );
	p += strlen(p);
	for (i = STSttVal; i < STNumVal; ++i)
	    if (!(stval[i].flags & (4|2)) && stval[i].value != 0) {
		sprintf(p, stval[i].fmt, stval[i].value);
		p += strlen(p);
	    }
	while (p > buff && *(p - 1) == ' ')
	    *--p = '\0';
	if (buff[0] != '\0') {
	    fprintf(tty, "\r%s", buff);
	    i = p - buff;
	    while (i < displen) {
		fputc(' ', tty);
		++i;
	    }
	    displen = p - buff;
	    fflush(stdout);
	    fflush(tty);
	}
    }
}

/* ask yes or no
 * return *yn = 'Y' or *yn = 'N'
 */
ST_Yes_or_No(mes, yn)
char *mes;
char *yn;
{
    char buff[BUFSIZ];

    if (displen) {
	fputc('\n', tty);
	displen = 0;
    }
    *yn = 'Y';
    return;
}

ST_Initialize()
{
    userabort = 0;
    setsignals();
}

ST_Terminate()
{
    dflsignals();
    terminated = 1;
}

setscreen()
{
    displen = 0;
}

unsetscreen()
{
    if (displen)
	fprintf(tty, "\n");
}

interactivemode()
{
    int prevc = -1;
    unsigned char c;
    
    setscreen();
    terminated = 0;

    while (!terminated) {
	while (!Async_BuffeR_Check(&c))
	    ;

	switch (c) {
	  case 'I':
	    if (prevc == ESC) {
		BP_Term_ESC_I(esci_response);
		prevc = -1;
	    } else
		prevc = c;
	    break;

	  case ENQ:
	    BP_Term_ENQ();
#ifdef DEBUG
	    fprintf(tty, "BP TERM ENQ\n");
#endif
	    
	    prevc = -1;
	    break;

	  case DLE:
	    BP_DLE_Seen();
#ifdef DEBUG
	    fprintf(tty, "BP DLE SEEN\n");
#endif
	    prevc = -1;
	    break;

	  default:
	    prevc = c;
	    break;
	}
    }

    unsetscreen();
}

static void
usage(mes)
char *mes;
{
    fprintf(tty, "%s\n", mes);
    exit(255);
}

main(argc, argv)
int argc;
char **argv;
{
    char *p;
    long atol();
    int bplus = 4;

    tty = stderr;
    ttyr = fopen( "/dev/tty" , "r" );

    dflsignals();

    BP_Use_File_Size = 0;
    BP_Auto_Resume = 1;

    interactive = 1;

    while (--argc)
	if (*(p = *++argv) == '-')
	    while (*++p)
		switch (*p) {
		case 'Q':
		    bplus = 1;
		    break;
		case 'E':
		    bplus = 2;
		    break;
		case 'D':
		    bplus =3;
		    break;
		case 'I':
		    interactive = 1;
		    bplus = 4;
		    break;
		case '.':
		    if (--argc < 1)
			usage("no iofd after -.");
		    infd = outfd = atoi(*++argv);
		    break;
		case 'S':
		    if (--argc < 1)
			usage("no filesize after -S");
		    BP_Use_File_Size = 1;
		    BP_File_Size = atol(*++argv);
		    break;
		case 'R':
		    BP_Auto_Resume = 1;
		    break;
		default:
		    sprintf(s128, "Unknown switch -%c", *p);
		    usage(s128);
		    break;
		}
	else
	    usage("don't specify other arguments");

    fprintf(tty, "%s\n", title);
    userabort = 0;

    cock2raw();

    switch (bplus) {
    case 1:
	BP_Term_ENQ();
	break;
    case 2:
	BP_Term_ESC_I(esci_response);
	break;
    case 3:
	setscreen();
	BP_DLE_Seen();
	unsetscreen();
	break;
    case 4:
	interactivemode();
	break;
    }

    raw2cock();

    fclose( ttyr );

    return 0;
}

/* ---- end of bp.c ---- */
