#include "qv.h"
#include "qvmsg.h"
#include <sys/times.h>
#include <sys/time.h>
#include <sys/timeb.h>

#ifdef V7
#include <sgtty.h>
#define OS "V7/BSD"
#endif

#ifndef OS
#ifndef USG
#define USG
#endif
#endif

#ifdef USG
#include <sys/ioctl.h>
#include <termio.h>
#define OS "SYS III/V"
#endif

extern char	*ttyname();
extern uint rxlen();
extern uint txlen();

#define	TRUE	1
#define	FALSE	0
#ifndef		PIPETEST
#define	ERROREND	{return -1;}
#else
#define	ERROREND	{;}			/* for debug with named pipe */
#endif
#define	RBUFSIZ		512

#ifdef ICANON
struct termio ttyorg, tty;	/* save original tty mode */
#else
struct sgttyb ttyorg, tty;	/* save original tty mode */
#endif

struct {
        ulong baudr;
        int speedcode;
} speeds[] = {
        110,    B110,
        300,    B300,
        600,    B600,
        1200,   B1200,
        2400,   B2400,
        4800,   B4800,
        9600,   B9600,
        19200,  EXTA,
        38400,  EXTB,
        0,
};
 
static ulong getspeed(code)
int code;
{
	register n;
 
	for (n=0; speeds[n].baudr; ++n)
		if (speeds[n].speedcode == code)
			return speeds[n].baudr;
	return 0;
}

int cominit( bps)
ulong	*bps;
{
	int		n;

#ifdef ICANON
	/* get original parameter */
	ioctl(0, TCGETA, &ttyorg);
	tty = ttyorg;
 
	tty.c_iflag = IGNBRK;
 
	/* No echo, crlf mapping, INTR, QUIT, delays, no erase/kill */ 
	tty.c_lflag &= ~(ECHO | ICANON | ISIG);
 
	tty.c_oflag = 0;	/* Transparent output */
 
	tty.c_cflag &= ~PARENB; /* Same baud rate, disable parity */    
	tty.c_cflag |= CS8 | CRTSCTS;	/* Set character size = 8 */
	tty.c_cc[VMIN] = 1; /* This many chars satisfies reads */
	tty.c_cc[VTIME] = 1;    /* or in this many tenths of seconds */ 
	ioctl(0, TCSETAW, &tty);
	if ( *bps == 0L)
		*bps = getspeed(tty.c_cflag & CBAUD);
#else
	/* get original parameter */
	ioctl(0, TIOCEXCL, 0);
	ioctl(0, TIOCGETP, &ttyorg);

        tty = ttyorg;
        tty.sg_flags |= RAW;
	tty.sg_flags &= ~ECHO;
	ioctl(0, TIOCSETP, &tty);
 
	ioctl(0, TIOCMSET, 0x0fff);
/*	n = 0;
	ioctl(0, TIOCSSOFTCAR, &n);*/

	/* get bps value */
	if ( *bps == 0L)
		*bps = getspeed(tty.sg_ospeed);
 
	n = 0;
	if (ioctl(0, TIOCFLUSH, &n) < 0)
		ERROREND
#endif
	return 0;
}

void comquit()
{
#ifdef ICANON
	ioctl(0, TCSBRK, 1);  /* Wait for output to drain */
	ioctl(0, TCFLSH, 1);  /* Flush input queue */
	ioctl(0, TCSETAW, &ttyorg);   /* Restore original modes */
	ioctl(0, TCXONC,1);   /* Restart output */
#else
	ioctl(0, TIOCSETP, &ttyorg);
	ioctl(0, TIOCNXCL, 0);
#endif
}

#ifdef ICANON
int	 comput(c)
int		c;
{
	char	data;
	data = c;
	if (write(1, &data, 1) != 1)
		return -1;
	return 0;
}

int	 comwrite(buf, size)
uchar	*buf;
int		size;
{
	if (write(1, buf, size) != size)
		return -1;
	return 0;
}
#else
int	 comput(c)
int		c;
{
	char	data;
	int	n;
	do	{
		ioctl( 0, TIOCMGET, &n);
		} while ( (n & TIOCM_CTS) == 0);
	data = c;
	if (write(1, &data, 1) != 1)
		return -1;
	return 0;
}

int	 comwrite(buf, size)
uchar	*buf;
int		size;
{
	int	n;
	while ( size-- > 0)
		{
		n = *buf++;
		comput( n);
		}
	return 0;
}
#endif
int comget(maxw)
int	maxw;
{
	int		rfd;
	uchar	c;
	struct timeval	tout;
	static uchar	rbuf[ RBUFSIZ ];
	static int		rnum = 0;
	static int		ridx;

	if (rnum == 0)
	{
		tout.tv_sec = maxw / 1000;
		tout.tv_usec = maxw % 1000;

		rfd = 1 << 0;
		if ((select(1, &rfd, NULL, NULL, &tout)) == 0)
			return -1;		/* wait data with timeout */
		if ((rnum = rxlen()) > RBUFSIZ)
			rnum = RBUFSIZ;
		rnum = read(0, rbuf, rnum);
		ridx = 0;
	}
	c = rbuf[ ridx++ ];
	if (ridx >= rnum)
		rnum = 0;
	return c;
}

ulong gettimer()
{
	struct timeb	t;
	static ulong	base = 0;

	if (base == 0)
	{
		ftime( &t );
		base = t.time % 100000L * 1000 + t.millitm;
		if (base == 0)
			base = 1;
	}
	ftime( &t );
	return t.time % 100000L * 1000 + t.millitm - base;
}

uint rxlen()
{
	long	l;

	ioctl(0, FIONREAD, &l);
	return (uint)l;
}

uint txlen()
{
	return 0;
}

void comRTSoff()
{
	/* nothing to do... OK? */
}

void comRTSon()
{
	/* nothing to do... OK? */
}

void strupr(str)
char	*str;
{
	for (; *str; str++)
		if (islower(*str))
			*str = toupper( *str );
}

ulong filelength(fd)
int		fd;
{
	struct stat	st;

	if (fstat(fd, &st) == 0)
		return st.st_size;
	else
		return 0L;
}
