/* $Id: pet_client.c,v 1.11 1996/02/07 01:13:29 kuroda Exp kuroda $ */
/*
ALL Rights Reserved, Copyright(c)FUJITSU LABORATORIES LTD. 1995
Privacy Enhanced Telnet

			 PERMISSION NOTICE
			 -----------------
			Wed, Dec 27th, 1995

		    FUJITSU LABORATORIES LTD.
  1015 Kamikodanaka, Nakahara-ku, Kawasaki-shi, Kanagawa-ken 211, Japan

Permission to use, copy, modify and distribute this Software (including its
relevand documentation) is hereby granted without fee under the
following terms and conditions:

1. Both of copyright notices in this Software and this permission
notice shall be duplicated in copies and modification of all or a part 
of this Software.

2. This Software may be used for the profit-making purposes, provided
that sale of this Software, whether incorporated in other products as
a part thereof or not, is prohibited and that the name FUJITSU
LABORATORIES LTD. may not be used externally in relation with the use
for such profit-making purposes without specific prior written
permission.

3. On publication or presentation in any form, of the results obtained 
by use of this Software, an explicit reference to this Software and
its copyright owner shall be made.

4. THIS SOFTWARE IS PROVIDED ``AS IS" AND FUJITSU LABORATORIES
LTD. MAKES NO WARRANTIES, EITHER EXPRESSED OR IMPLIED, AS TO ANY
MATTER WHATSOEVER REGARDING TO THIS SOFTWARE, INCLUDING, WITHOUT
LIMITATION, WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY
PARTICULAR PUPRPOSE.
IN NO EVENT SHALL FUJITSU LABORATORIES LTD. BE LIABLE FOR ANY CLAIM
FROM A THIRD PARTY, OR SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
ARISING OUT OF OR IN CONNECTION WITH THE USE, COPYING, MODIFICATION OR 
DISTRIBUTION OF THIS SOFTWARE.

5. this Software includes the following free software, and such free
software shall be used in accordance with the permission notices
attached to each such free software:


MANE OF FREE SOFTWARE	NAME OF DIRECTORY	NAME OF COPYRIGHT OWNER
UCB Telnet              telnet			UCB (1)        
UCB Telnetd             telnetd			UCB            
UCB libtelnet           libtelnet		UCB            
GNU DES Libray          DES			Eric Young     
OSISEC RSA Libray       RSA			Peter Williams 
OSISEC MD Libray        MD			Peter Williams 
transit		        transit			ASCII Corporation
bplus               	bplus			K.Suzuki(katsu@wide.ad.jp)
						T.Taniguchi(tani@wide.ad.jp)
zmodem              	zmodem			Omen Technology Inc.
telnetx                 telnet/telnetx.c	UEC (2)      

PET3 Library            pet			FUJITSU LABORATORIES LTD.

	(1) UCB: The University of California, Berkeley School
	(2) UEC: The University of Electro-Communications


*/


#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>

#include "../telnet/ring.h"
#include "../telnet/externs.h"
#include "../telnet/defines.h"
#include "../arpa/telnet.h"

#include "../DES/des.h"
#include "../RSA/numalgs.h"
#include "../RSA/rsa.h"

#include "pet.h"
#include "pet2.h"
#include "cert.h"


FILE *fin, *fout;
int verbose = 0;

int pet_mode;

char *seclevel_table[] = {
    "110000",
    "112040",
    "022040",
    "002305",
    "000040",
    "000045"
};
#define KEYSIZB 1024

extern  void    initcfb();
extern  void    initcfb2();

extern  void    initcfb3();
extern  int     encfb3();
extern  int     decfb3();

des_key_schedule kss;
des_key_schedule ks;
int fbsize = 8;
static char     key3[KEYSIZB+1];
static char     rrbuf[BUFSIZ];
static char     ccbuf[BUFSIZ];
static char     rbuf[BUFSIZ];
static char     cbuf[BUFSIZ];
static char     key[KEYSIZB+1];

unsigned char dek1[1024], dek11[1024], dek22[1024], dek2[1024], dek3[1024];

static int      binptr2 = 0, nbytes2 = 0;

BER *getber3();
BER *getber2();
BER *getber();

BER *server;
char cseed[1024];
struct  RSAPublicKey    spub;
struct  RSAParameters   sparams;

pet_client_do_level2(){
    char *level2 = "level2\0";

    pet_write_data(DO, level2);

}

pet_client_vfy_ssig(ssig,count)
char *ssig;
int count;
{
    static unsigned char bin1[1024];
    static unsigned buf1[1024];
    static  unsigned  *w, *ww;
    char *errstr, cseed2[1024];
    unsigned char seed2[1024];

    str2char(ssig, bin1);
    char2word(bin1, buf1, (sparams.len*2));
    rsa_encrypt(buf1, &ww, spub, sparams);
    word2char(ww, bin1, sparams.len);
    padout(bin1, seed2, 8);
    char2str(seed2, cseed2, 8);
/*    fprintf(stderr, "sseed2 = %s\n", cseed2);*/
    if(strcmp(cseed, cseed2)){
        fprintf(stderr, "Server's authentication is Invalid\r\n");
        fprintf(stderr, "Connection Closed.\r\n");
	ring_consumed(&netiring, count);
	Exit(0);
/*        exit(0);*/
    }else{
        fprintf(stderr, "Server's authentication is Valid\r\n");
    }
}


pet_client_vfy_scert(scert,count)
char *scert;
int count;
{
    FILE *fp;
    char buff[1024];
    BER *myissuer;
    struct  RSASecretKey     *skey = NULL, *mykey = NULL;
    struct  RSAPublicKey    ipub, mypub;
    struct  RSAParameters   iparms, myparams, my_params;
    char *errstr, c;
    unsigned char seed[1024];
    char myhome[128];
    char myissuerfile[128], mykeyfile[128];

    strcpy(myhome, getenv("HOME"));
    strcpy(myissuerfile, myhome);
    strcat(myissuerfile, "/.issuer");
    if((fp = fopen(myissuerfile, "r")) == NULL){
        fprintf(stderr, "Bad \"%s\" file\n", myissuerfile);
	ring_consumed(&netiring, count);
	Exit(0);
    }
    myissuer = getber3(fp);
/*    prnname(C_SN(myissuer));*/
    getpubkey(C_PK(myissuer), &ipub, &iparms);
    fclose(fp);

    server = getber2(scert);
    fprintf(stderr, "Remote server name is\n");
    prnname(C_SN(server));
    getpubkey(C_PK(server), &spub, &sparams);
    errstr = vfysig(server, &ipub, &iparms);
    if(errstr){
        fprintf(stderr, "Server's certificate is invalid\r\n");
        fprintf(stderr, "Connection Closed.\r\n");
	ring_consumed(&netiring, count);
	Exit(0);
/*        exit(0);*/
    }else{
        fprintf(stderr, "Server's certificate is valid\r\n");
    }



    srandom((int)time(NULL) + getpid());
    strcpy(mykeyfile, myhome);
    strcat(mykeyfile, "/.key");
    if((fp = fopen(mykeyfile, "r")) == NULL){
        fprintf(stderr, "Bad \"%s\" file\n", mykeyfile);
	ring_consumed(&netiring, count);
	Exit(0);
    }
    fclose(fp);
    rsa_read_secret_key(mykeyfile, &mykey, &my_params);
    makedek(seed);
    char2str(seed, cseed, 8);

    pet_write_data(CSEED, cseed);
/*    fprintf(stderr, "cseed = %s\n", cseed);*/


}


pet_client_sseed_result(result,count)
char *result;
int count;
{
    char *valid = "valid\0";

    if(!strcmp(result, valid))
        fprintf(stderr, "Your authentication is valid.\r\n");
    else{
        fprintf(stderr, "Your authentication is invalid.\r\n");
        fprintf(stderr, "Connection Closed.\r\n");
	ring_consumed(&netiring, count);
	Exit(0);
/*        exit(0);*/
    }
}


pet_client_sseed_sig(sseed,count)
char *sseed;
int count;
{
    FILE *fp;
    unsigned char seed[1024];
    struct  RSASecretKey     *mykey = NULL;
    struct  RSAPublicKey    mypub;
    struct  RSAParameters   myparams, my_params;
    BER *mycert;
    static unsigned char bin1[1024];
    char *errstr, str1[1024], str2[1024], padstr[1024], str3[1024];
    static unsigned buf1[1024];
    char myhome[128], mycertfile[128], mykeyfile[128];
    static  unsigned  *w;

    strcpy(myhome, getenv("HOME"));
    strcpy(mycertfile, myhome);
    strcat(mycertfile, "/.cert");
    if((fp = fopen(mycertfile, "r")) == NULL){
        fprintf(stderr, "Bad \"%s\" file\n", mycertfile);
	ring_consumed(&netiring, count);
	Exit(0);
/*        exit(1);*/
    }
    mycert = getber3(fp);
    fclose(fp);
    getpubkey(C_PK(mycert), &mypub, &myparams);

    strcpy(mykeyfile, myhome);
    strcat(mykeyfile, "/.key");
    if((fp = fopen(mykeyfile, "r")) == NULL){
        fprintf(stderr, "Bad \"%s\" file\n", mykeyfile);
	ring_consumed(&netiring, count);
	Exit(0);
    }
    fclose(fp);
    rsa_read_secret_key(mykeyfile, &mykey, &my_params);

    str2char(sseed, seed);
    pad(1, seed, 8, bin1, (myparams.len*2));
    char2str(bin1, str3, (myparams.len *2));
    char2word(bin1, buf1, (myparams.len*2));
    rsa_decrypt(buf1, &w, mykey, myparams);
    word2char(w, bin1, myparams.len);
    char2str(bin1, padstr, (myparams.len *2));

    pet_write_data(SIGSSEED, padstr);

}

pet_client_rcv_ccertresult(result,count)
char *result;
int count;
{
    char *valid = "valid";

    if(!strcmp(result, valid))
        fprintf(stderr, "Your certificate is valid.\r\n");
    else{
        fprintf(stderr, "Your certificate is invalid.\r\n");
        fprintf(stderr, "Connection Closed.\r\n");
	ring_consumed(&netiring, count);
	Exit(0);
/*        exit(0);*/
    }

}


pet_client_send_cert(count)
int count;
{
    FILE *fp;
    char buff[1024];
    BER *server, *mycert, *myissuer;
    struct  RSASecretKey     *skey = NULL, *mykey = NULL;
    struct  RSAPublicKey    spub, ipub, mypub;
    struct  RSAParameters   sparams, iparms, myparams, my_params;
    char *errstr, str1[1024], str2[1024], padstr[1024], str3[1024];
    unsigned char seed[1024], seed2[1024];
    static unsigned char bin1[1024];
    static unsigned buf1[1024];
    static  unsigned  *w, *ww;
    int len, i;
    char c, tmp[10], auth_ok[128],sig_ok[128];
    char myhome[128], mycertfile[128], mykeyfile[128];
    char *valid = "valid", *invalid = "invalid";

    /* read my certificate */
    strcpy(myhome, getenv("HOME"));
    strcpy(mycertfile, myhome);
    strcat(mycertfile, "/.cert");
    if((fp = fopen(mycertfile, "r")) == NULL){
        fprintf(stderr, "Bad \"%s\" file\n", mycertfile);
	ring_consumed(&netiring, count);
	Exit(0);

/*        exit(1);*/
    }
    mycert = getber3(fp);
    fclose(fp);
    getpubkey(C_PK(mycert), &mypub, &myparams);
    strcpy(mykeyfile, myhome);
    strcat(mykeyfile, "/.key");
    if((fp = fopen(mykeyfile, "r")) == NULL){
        fprintf(stderr, "Bad \"%s\" file\n", mykeyfile);
	ring_consumed(&netiring, count);
	Exit(0);
    }
    fclose(fp);
    rsa_read_secret_key(mykeyfile, &mykey, &my_params);
    fprintf(stderr, "Your name is\n");
    prnname(C_SN(mycert));

    /* send client certificate to server */
    if((fp = fopen(mycertfile, "r")) == NULL){
        fprintf(stderr, "Bad \"%s\" file\n", mycertfile);
	ring_consumed(&netiring, count);
	Exit(0);

/*        exit(1);*/
    }
    for(i=0; (c = fgetc(fp)) != EOF; i++){
        buff[i] = c;
    }
    fclose(fp);

    pet_write_data(CCERT, buff);


}


pet_client_decdek2(encdek2,count)
char *encdek2;
int count;
{
    int i;

    str2char(encdek2, dek3);

    for(i=0; i<DEKSIZE; ++i){
        dek2[i] = decfb3(dek3[i]);
    }
    char2str(dek2, dek22, DEKSIZE);

/*    fprintf(stderr, "dek22(c) = %s\n", dek22);*/

    initcfb(dek22);
    initcfb2(dek22);

/*    pet_mode = 1;*/

}


pet_client_signature(pubkey,count)
char *pubkey;
int count;
{
    struct RSAPublicKey *pub;
    struct RSAParameters *parms;
    char data[1024];
    static  unsigned char   bin[BUFSIZ];

    fprintf(stderr, "Now, exchanging a data encryption key. Please wait.\r\n");
    set_pubkey(pubkey, &pub, &parms);

    srandom((int)time(NULL) + getpid());
    makedek(dek1);
    keyinfo(dek1, pub, parms, bin);
    char2str(bin, data, parms->len * 2);

    pet_write_data(DEK1, data);

    char2str(dek1, dek11, DEKSIZE);
    initcfb3(dek11);


}


pet_client_req_level(rlevel,count)
char *rlevel;
int count;
{

    if(strcmp(creq_level, "") == 0){
        if(strcmp(rlevel, "1") == 0){
            pet_mode = 0;
        }
        fprintf(stderr, "You don't set your security level.\n");
        fprintf(stderr, "Now, I set your security level to remote server's security level %s.\r\n", rlevel);
        strcpy(creq_level, rlevel);
/*	sreq_level[0] = rlevel[0];
	sreq_level[1] = '\0';*/
    }else{
        printf("You are requesting security level %s.\n", 
            creq_level);
    }

    fprintf(stderr, "Remote server is requesting security level %s.\n", 
            rlevel);

/*    creal_level = seclevel_table[atoi(creq_level)][atoi(rlevel)];*/
    creal_level = seclevel_table[atoi(creq_level)][atoi(rlevel)];

    fprintf(stderr, "We are going to talk on security level %c.\r\n",
	    creal_level);

    pet_write_data(CREQLEVEL, creq_level);
}



pet_write_data(type, data)
char type;
char *data;
{
    char send_data[1024];
    int i, j, len;
    FILE *fp;
    
    send_data[0] = IAC;
    send_data[1] = PET;
    send_data[2] = type;

    len = strlen(data);
    for(i=3, j=0; j<len+1; i++,j++)
        send_data[i] = data[j];

    len = strlen(send_data);

    write(net, send_data, len+1);

}

set_pubkey(cpub, out, parms)
char cpub[];
struct RSAPublicKey **out;
struct RSAParameters **parms;
{
    struct RSAPublicKey *result;
    struct RSAParameters *newparms;
    int len;

    result = (struct RSAPublicKey *) calloc((unsigned)1, sizeof(*result));
    if (result == (struct RSAPublicKey *) 0)
        return (NOTOK);

    newparms = (struct RSAParameters *) calloc((unsigned)1, 
                                               sizeof(*newparms));
    if (newparms == (struct RSAParameters *) 0)
                return (NOTOK);

    newparms->len = strlen(cpub) * 4;
    newparms->shift = 16 - newparms->len % 16;
    if (newparms->shift == 16)
        newparms->shift = 0;
    newparms->len = (newparms->len + 15) / 16;

    result->n = (unsigned *) calloc((unsigned)newparms->len, 
                                    sizeof(unsigned));
    if (result->n == NULLENCRYPTED)
        return (NOTOK);

    hexin2(cpub, result->n, newparms->len - 1);
    result->e = 65537;

    (*out) = result;
    (*parms) = newparms;

    return (OK);


}

hexdump2(indata, outdata, l)
unsigned indata[];
char outdata[];
int             l;
{
    int             i, j, len;
    unsigned        n;
    char tmp[10];
    FILE *fp;

    for (i = l; i >= LSW; i--) {
        n = indata[i];
        for (j = 0; j < BITS; j = j + 4) {
            sprintf(tmp, "%x", n >> HNIB);
            strcat(outdata,tmp);
            n = (n << 4) & MASK;
        }
    }

}


padout(in, out, size)
unsigned char   *in, *out;
int             size;
{
    ++in; /* for the precedig 00 */
    while(*in++) ;
    while(size-- > 0){
        *out++ = *in++;
    }
}

hexin2(in, a, l)
char in[];
unsigned a[];
int l;
{
    int             i;
    int             j;
    int k;
    int             n;
    char            c;

    k=0;
    for (i = l; i >= 0; i--) {
        a[i] = 0;
        for (j = 0; j < NPW; j++) {
            /*                        c = getc(f);*/
            c = in[k++];
            n = (int) c - (int) '0';
            if (c > '9')
                n = (int) c - (int) 'a' + 10;
            a[i] = (a[i] << 4) + n;
        }
    }
}

keyinfo(dek, pub, parms, bin)
unsigned char   dek[];
struct RSAPublicKey     *pub;
struct RSAParameters    *parms;
unsigned char  bin[];
{
    static  unsigned        buf[BUFSIZ], *w;
    int     bytes, i;

    bytes = parms->len * 2;
    pad(2, dek, DEKSIZE, bin, bytes);

    char2word(bin, buf, bytes); 
    rsa_encrypt(buf, &w, pub, parms);
    word2char(w, bin, parms->len);

}


getdek(bin, dek, mykey, parms)
unsigned char  bin[];
unsigned char   dek[];
struct  RSASecretKey    *mykey;
struct  RSAParameters   *parms;
{
    unsigned        *cert;
    static unsigned buf[BUFSIZ];
    int             i, c;

    char2word(bin, buf, (parms->len *2)); 
    rsa_decrypt(buf, &cert, mykey, parms);
    word2char(cert, bin, parms->len);

    padout(bin, dek, 8); 

    return(0);

}

makedek(dek)
unsigned char   dek[];
{
    int     i, j, s;

    for(i = 0; i < DEKSIZE; ++i){
        dek[i] = random() & 0x7f;
        for(j = 0, s = 0; j < 7; ++j){
            s += (dek[i] >> j) & 1;
        }
        if(s & 1){
            dek[i] |= 0x80;
        }
    }
}

char2str(inbuf, outbuf, len)
unsigned char inbuf[];
char outbuf[];
int len;
{
    int i;
    char tmp[10];
    FILE *fp;

    outbuf[0] = '\0';
    for(i=0; i<len; i++){
        sprintf(tmp, "%02x", inbuf[i]);
        strcat(outbuf, tmp);
    }

}

str2char(inbuf, outbuf)
char inbuf[];
unsigned char outbuf[];
{
    int i, j, n, nn, len;

    len = strlen(inbuf);
    for(i=0,j=0; i<len; ++i){
        n = (int)inbuf[i] - (int)'0';
        if(inbuf[i] > '9')
            n = (int)inbuf[i] - (int)'a' +10;
        n = n*16; ++i;
        nn = (int)inbuf[i] - (int)'0';
        if(inbuf[i] > '9')
            nn = (int)inbuf[i] - (int)'a' +10;
        n = n + nn;
        outbuf[j++] = n;
    }

}



char2word(c, w, csize)
unsigned char   c[];
unsigned int    w[];
int             csize;
{
    int     i, j;

    for(i = csize -1, j = 0; i >= 0; i -= 2){
        w[j++] = c[i] + (c[i-1] << 8);
    }
}

word2char(w, c, wsize)
unsigned int    w[];
unsigned char   c[];
int             wsize;
{
    int     i, j;

    for(i = wsize -1, j = 0; i >= 0; --i){
        c[j++] = w[i] >> 8;
        c[j++] = w[i] & 0xff;
    }
}

/*  padding according to PKCS #1 */

pad(type, dek, msize, pad, psize)
unsigned char   *dek, *pad, type;
int             msize, psize;
{
    int     i, mp;
    mp = psize - msize -1;

    pad[0] = 0;
    pad[1] = type;
    for(i = 2; i < mp; ++i){
        do{
            pad[i] = (type == 1)? 0xff: (random()) & 0xff;
        }while(pad[i] == 0);
    }
    pad[mp] = 0;
    bcopy(dek, &pad[mp +1], msize);
}

void initcfb3(key3)
char* key3;
{
        static des_cblock kkk;

        des_string_to_key(key3,kkk);
        des_set_key(kkk,kss);
        bzero(key3,sizeof(key3));
        bzero(kkk,sizeof(kkk));
        bzero(rrbuf,sizeof(rrbuf));
        bzero(ccbuf,sizeof(ccbuf));

}

int encfb3(c)
int c;
{
        static int      i = 8;

        if(i == 8){
                des_ecb_encrypt(rrbuf, ccbuf, kss, DES_ENCRYPT);
                i = 0;
        }
        rrbuf[i] = c^ccbuf[i];
        return(rrbuf[i++]);
}

int decfb3(c)
int c;
{
        static int      i = 8;

        if(i == 8){
                des_ecb_encrypt(rrbuf, ccbuf, kss, DES_ENCRYPT);
                i = 0;
        }

        rrbuf[i] = c;
        return(c^ccbuf[i++]);    
}


BER*
getber2(buff)
unsigned char   buff[];
{
    unsigned char   buf[BUFSIZ];

    bufreadfield(buff, buf);
    return(deber(buf));
}

BER*
getber3(fp)
FILE *fp;
{
    unsigned char   buf[BUFSIZ];

    bufreadfield2(fp, buf);
    return(deber(buf));
}

bufreadfield(buff, field)
unsigned char   buff[];
unsigned char   field[];
{
/*    static unsigned char        buff[BUFSIZ];*/
    static unsigned char        bin[BUFSIZ];
    static unsigned char        line[BUFSIZ];
    int   len, i, j;
    int binptr = 0, nbytes = 0;
    char        *cp, *tp;
    int c;

/*    for(i=0; (c = fgetc(f)) != EOF; i++){
        buff[i] = c;
    }*/

    len = strlen(buff);
/*    fprintf(stderr, "len = %d\n", len);*/
/*    while((cp = fgets(line, BUFSIZ, f)) != NULL && line[0] == ' '){*/
    for(i=0; i<len; i++){
        j=0;
        while(buff[i] != '\n'){
            line[j++] = buff[i++];
/*          fprintf(stderr, "%c", line[j]);*/
        }
        c = strlen(line);
        /* if(verbose) printf("[%d]%s", c, line); */
/*      printf("[%d]%s", c, line); */
        nbytes += prdecode(line, &field[nbytes], (c-1)*6/8);
        if(tp = strpbrk(line, "=,")){
            if(verbose){
                fprintf(stderr, "\tNo ungetc,(%d) \n", nbytes);
            }
            if(tp = strpbrk(tp, ",")){
                if(verbose) fprintf(stderr, "ungets(%s)\n", tp + 1);
                ungets(tp + 1);
            }
            return(nbytes);
        }
    }
    if(cp != NULL){
        ungets(line);
    }
    if(verbose) fprintf(stderr, "(%d)\n", nbytes);
    return(nbytes);
}



bufreadfield2(f, field)
FILE *f;
unsigned char   field[];
{
    static unsigned char        buff[BUFSIZ];
    static unsigned char        bin[BUFSIZ];
    static unsigned char        line[BUFSIZ];
    int   len, i, j;
    int binptr = 0, nbytes = 0;
    char        *cp, *tp;
    int c;

/*    fprintf(stderr, "%s\n", buff);*/
    for(i=0; (c = fgetc(f)) != EOF; i++){
        buff[i] = c;
    }

    len = strlen(buff);
/*    fprintf(stderr, "len = %d\n", len);*/
/*    while((cp = fgets(line, BUFSIZ, f)) != NULL && line[0] == ' '){*/
    for(i=0; i<len; i++){
        j=0;
        while(buff[i] != '\n'){
            line[j++] = buff[i++];
/*          fprintf(stderr, "%c", line[j]);*/
        }
        c = strlen(line);
        /* if(verbose) printf("[%d]%s", c, line); */
/*      printf("[%d]%s", c, line); */
        nbytes += prdecode(line, &field[nbytes], (c-1)*6/8);
        if(tp = strpbrk(line, "=,")){
            if(verbose){
                fprintf(stderr, "\tNo ungetc,(%d) \n", nbytes);
            }
            if(tp = strpbrk(tp, ",")){
                if(verbose) fprintf(stderr, "ungets(%s)\n", tp + 1);
                ungets(tp + 1);
            }
            return(nbytes);
        }
    }
    if(cp != NULL){
        ungets(line);
    }
    if(verbose) fprintf(stderr, "(%d)\n", nbytes);
    return(nbytes);
}
