/*
 * $Id: qvrecv.c,v 1.2 1994/09/26 18:33:41 tani Exp $
 */

#include "qv.h"
#include "qvmsg.h"
static int      FileNumber = 1;

static int RecvOneBlock(xb, timer)
	XBLOCK         *xb;
	ulong           timer;
{
  int             size, c;
  unsigned char   ChkSum, sum;
  
  for (sum = size = 0; size < BLOCKSIZE; size++) {
    c = comget(timer);
    if (c == TIMEOUT) {
      fprintf(stderr, E_TimeOut);
      return (ERROR);
    }
    sum += c;
    xb->Body[size] = c;
  }
  if ((~xb->Num1 & 0x00ff) != xb->Num2) {
    sum += xb->Header;
    sum += xb->Num1;
    sum += xb->Num2;
  }
  ChkSum = (c = comget(timer));
  if (c == TIMEOUT) {
    fprintf(stderr, E_TimeOut);
    return (ERROR);
  }
  XferTotalBlock++;
  if (ChkSum != sum) {
    fprintf(stderr, E_CheckSum);
    return (ERROR);
  }
  return (NORMAL);
}

static int NegoSubPhase1(xparam)
	XPARAM         *xparam;
{
  int             Retry;
  ulong           timer;
  int             c;
  XBLOCK          xb;
  
  for (Retry = 0; Retry < MaxRetry; Retry++) {
    SendCtrl(NAK);
    timer = SetTimer(RTimer);
    while ((c = comget(timer)) != TIMEOUT) {
      if ((c == CAN) && (NothingArrive())) {
	fprintf(stderr, E_Canceled);
	return (FATAL);
      }
      if (c == SOH) {
	xb.Header = c;
	if ((c = comget(timer)) == TIMEOUT)
	  break;
	xb.Num1 = c;
	if ((c = comget(timer)) == TIMEOUT)
	  break;
	xb.Num2 = c;
	if (xb.Num1 != 0 || xb.Num2 != SINIT)
	  continue;
	if (RecvOneBlock(&xb, timer) == NORMAL) {
	  memcpy(xparam, xb.Body, sizeof(XPARAM));
	    return (NORMAL);
	}
      }
    }
  }
  fprintf(stderr, E_NEGO1Fail);
  SendCtrl(CAN);
  return (FATAL);
}

static SendAnswerBlock(AnswerID, SeqNum, body, size)
	int             AnswerID;
	int             SeqNum;	
	char           *body;	
	int             size;  
{
  int             sum;
  
  comput(STX);
  comput(AnswerID);
  comput(0x80 | (SeqNum & 0x7f));
  if (AnswerID == VACK || AnswerID == VNAK) {
    size = 0;
  }
  for (sum = STX + AnswerID + SeqNum; size > 0; size--, body++) {
    comput(*body);
    sum += *body;
  }
  comput(0x80 | (sum & 0x7f));
  comput(CR);
}

static int NegoSubPhase2(xparam)
	XPARAM         *xparam;	
{
  int             Retry;
  ulong           timer;
  int             c;
  XBLOCK          xb;
  
  for (Retry = 0; Retry < MaxRetry; Retry++) {
    xparam->Version = '1';
    SendAnswerBlock(RINIT, 0, xparam, strlen((char *) xparam));
      timer = SetTimer(RTimer);
    while ((c = comget(timer)) != TIMEOUT) {
      if ((c == CAN) && (NothingArrive())) {
	fprintf(stderr, E_Canceled);
	return (FATAL);
      }
      if (c == SOH) {
	xb.Header = c;
	if ((c = comget(timer)) == TIMEOUT)
	  break;
	xb.Num1 = c;
	if ((c = comget(timer)) == TIMEOUT)
	  break;
	xb.Num2 = c;
	if (xb.Num1 == 0 && xb.Num2 == VFILE) {
	  if (RecvOneBlock(&xb, timer) != NORMAL)
	    break;
	  if (!isdigit(xb.Body[0]) || !isdigit(xb.Body[1])) {
	    fprintf(stderr, E_VFILE, xb.Body);
	    goto error;
	  }
	  SetFileInfo(xb.Body);
	  if (XferFileNum != FileNumber) {
	    fprintf(stderr, E_VFILE, xb.Body);
	    goto error;
	  }
	  LastBlockNo = (XferSize + 127) / 128;
	  return (NORMAL);
	}
      }
    }
  }
  error:
  fprintf(stderr, E_NEGO2Fail);
  SendCtrl(CAN);
  return (FATAL);
}

static int SkipSubPhase()
{
  int             Retry;
  ulong           timer;
  int             c;
  XBLOCK          xb;
  char            Fnum[3];
  
  sprintf(Fnum, "%02d", XferFileNum);
  for (Retry = 0; Retry < MaxRetry; Retry++) {
    SendAnswerBlock(VSKIP, 0, Fnum, 2);
      timer = SetTimer(RTimer);
    while ((c = comget(timer)) != TIMEOUT) {
      if (c == EOT && NothingArrive()) {
	SendCtrl(ACK);
	return (SKIP);
      }
      if (c == CAN && NothingArrive()) {
	fprintf(stderr, E_Canceled);
	return (FATAL);
      }
      if (c == SOH) {
	xb.Header = c;
	if ((c = comget(timer)) == TIMEOUT)
	  break;
	xb.Num1 = c;
	if ((c = comget(timer)) == TIMEOUT)
	  break;
	xb.Num2 = c;
	if (xb.Num1 == 0 && xb.Num2 == VFILE) {
	  if (RecvOneBlock(&xb, timer) != NORMAL)
	    break;
	  if (!isdigit(xb.Body[0]) || !isdigit(xb.Body[1])) {
	    fprintf(stderr, E_VFILE, xb.Body);
	    goto error;
	  }
	  SetFileInfo(xb.Body);
	  if (XferFileNum == FileNumber + 1) {
	    FileNumber++;
	    LastBlockNo = (XferSize + 127) / 128;
	    return (NEXT);
	  }
	  if (XferFileNum != FileNumber) {
	    fprintf(stderr, E_VFILE, xb.Body);
	    goto error;
	  }
	}
      }
    }
  }
  error:
  fprintf(stderr, E_NEGO2Fail);
  SendCtrl(CAN);
  return (FATAL);
}

static int RecvStart(path)
	char           *path;		
{
  int             Retry;
  ulong           timer;
  int             c, sts;
  XBLOCK          xb;
  char            Offset[10];
  char            fname[128];
  
  DispReceiveFile(XferFile, XferSize, LastBlockNo);
  if ((sts = OpenRecvFile(path)) == FATAL) {
    SendCtrl(CAN);
    fprintf(stderr, E_RecvFileOpen, XferFile);
    return (FATAL);
  }
  if (sts == ERROR) {
    fprintf(stderr, E_FileSkip, XferFile);
    if (SkipFunc == 'Y')
      return (SkipSubPhase());
    SendCtrl(CAN);
    return (FATAL);
  }
  if (StartBlockNo == 1) {
    Offset[0] = '\0';
  } else {
    sprintf(Offset, "%d", StartBlockNo - 1);
  }
  for (Retry = 0; Retry < MaxRetry; Retry++) {
      SendAnswerBlock(VRPOS, 0, Offset, strlen(Offset));
    timer = SetTimer(RTimer);
    while ((c = comget(timer)) != TIMEOUT) {
      if ((c == CAN) && (NothingArrive())) {
	fprintf(stderr, E_Canceled);
	return (FATAL);
      }
      if (c == SOH) {
	return (NORMAL);
      }
    }
  }
  fprintf(stderr, E_RecvStartFail);
  SendCtrl(CAN);
  return (FATAL);
}

static int RecvSeqNum(seqnum, timer)
	int            *seqnum;
	ulong           timer;
{
  int             c;
  unsigned char   Num, CNum;
  
  Num = (c = comget(timer));
  if (c == TIMEOUT) {
    fprintf(stderr, E_TimeOut);
    return (ERROR);
  }
  CNum = (c = comget(timer));
  if (c == TIMEOUT) {
    fprintf(stderr, E_TimeOut);
    return (ERROR);
  }
  if (CNum != (~Num & 0x00ff)) {
    if (Num == 0 && CNum == VENQ) {
      return (VENQ);
    } else {
      fprintf(stderr, E_BadSeqNo);
      return (FATAL);
    }
  }
  *seqnum = Num;
  return (NORMAL);
}

static SendNak(seqnum, retry)
	int             seqnum;
	int            *retry;
{
  
  *retry += 1;
  if (*retry > RCount) {
    fprintf(stderr, E_RetryOver);
    return (FATAL);
  }
  SendAnswerBlock(VNAK, seqnum, 0, 0);
  fprintf(stderr, E_ReXfer, seqnum);
  XferRetry++;
  return (NORMAL);
}

static WaitNextBlock(timer, RecvSeqNo, RetryCounter)
	ulong          *timer;
	int             RecvSeqNo;
	int            *RetryCounter;
{
  int             c;
  
  do {
    c = comget(*timer);
    if ((c == CAN) && NothingArrive()) {
      fprintf(stderr, E_Canceled);
      return (CAN);
    }
    if (c == TIMEOUT) {
      fprintf(stderr, E_TimeOut);
      if (SendNak(RecvSeqNo, RetryCounter) != NORMAL) {
	return (FATAL);
      }
      ResetTimer(*timer, RTimer);
    }
  } while (c != SOH);
  return (SOH);
}

static int RecvFile(LastSeqNo)
	int            *LastSeqNo;
{
  int             RecvSeqNo;
  ulong           timer;
  int             RetryCounter = 0;
  int             seqnum, sts;
  XBLOCK          xb;
  
  XferStart = gettimer();
  XferBlock = 0;
  XferTotalBlock = 0;
  XferRetry = 0;
  timer = SetTimer(RTimer);
  RecvSeqNo = 1;
  do {
    sts = RecvSeqNum(&seqnum, timer);
    if (sts == NORMAL) {
      xb.Num1 = seqnum;
      xb.Num2 = (uchar) ~ xb.Num1;
      sts = RecvOneBlock(&xb, timer);
      if (seqnum == RecvSeqNo) {
	if (sts == NORMAL) {
	  if (WriteOneBlock(xb.Body) != NORMAL) {
	    goto error;
	  }
	  XferBlock += 1;
	  RetryCounter = 0;
	  if ((XferBlock % AckRate) == 0) {
	    SendAnswerBlock(VACK, RecvSeqNo, 0, 0);
	    DisplayStatus();
	  }
	  RecvSeqNo = (RecvSeqNo + 1) & 0xFF;
	} else {
	  if (SendNak(RecvSeqNo, &RetryCounter) != NORMAL) {
	    goto error;
	  }
	}
      } else {
	if (RetryCounter == 0) {
	  int             diff;
				
	  diff = seqnum - RecvSeqNo;
	  if (diff > 0 && diff < WindowSizeMax) {
	    fprintf(stderr, E_IllSeqNo, seqnum, RecvSeqNo);
	    SendNak(RecvSeqNo, &RetryCounter);
	  }
	}
      }
      ResetTimer(timer, RTimer);
    } else {
      if (sts == VENQ) {
	xb.Header = SOH;
	xb.Num1 = 0;
	xb.Num2 = VENQ;
	sts = RecvOneBlock(&xb, timer);
	if (sts == NORMAL) {
	  *LastSeqNo = (RecvSeqNo - 1) & 0xFF;
	  if (xb.Body[0] == *LastSeqNo) {
	    if (xb.Body[1] == Eof) {
	      XferTotalBlock--;
	      XferFinish = gettimer();
	      CloseRecvFile();
	      return (NORMAL);
	    } else {
	      if (xb.Body[1] < '1' || xb.Body[1] > '9') {
		fprintf(stderr, E_VenqFunc);
		goto error;
	      }
	    }
	  } else {
	      sts = ERROR;
	  }
	}
      }
      if (sts == ERROR) {
	if (SendNak(RecvSeqNo, &RetryCounter) != NORMAL) {
	  goto error;
	}
	ResetTimer(timer, RTimer);
      }
    }
  } while ((sts = WaitNextBlock(&timer, RecvSeqNo, &RetryCounter)) == SOH);
  
  error:
  SendCtrl(CAN);
  fprintf(stderr, E_ABEND, StartBlockNo + XferBlock - 1);
  return (FATAL);
}

static DecideXparameter(xparam)
	XPARAM         *xparam;
{
  XPARAM          Rxpara;
  int             Wsize;
  static int      AckRateTable[] =
    {1, 1, 1, 2, 2, 2, 3, 3, 3, 5, 5, 5, 5, 5, 5, 7, 7, 7, 7, 7, 7, 11, 11, 11, 11, 11, 11, 11, 11, 11};
  
  memcpy(&Rxpara, xparam, sizeof(Rxpara));
  if (!isdigit(xparam->WindowSIZE[0]))
    goto error;
  Wsize = (xparam->WindowSIZE[0] - '0') * 10;
  if (!isdigit(xparam->WindowSIZE[1]))
    goto error;
  Wsize = Wsize + (xparam->WindowSIZE[1] - '0');
  if (Wsize > WindowSizeMax) {
    xparam->WindowSIZE[0] = (WindowSizeMax / 10) + '0';
    xparam->WindowSIZE[1] = (WindowSizeMax % 10) + '0';
  } else {
    WindowSizeMax = Wsize;
  }
  switch (xparam->OptionNumber) {
    case '0':
    xparam->OptionFunc1 = '\0';
    break;
    case '1':
    if (xparam->OptionFunc1 == '1') {
      xparam->Last = '\0';
      SkipFunc = 'Y';
    } else {
      goto error;
    }
    break;
    default:
    goto error;
  }
  AckRate = AckRateTable[WindowSizeMax - 1];
    return (NORMAL);
  
  error:
  fprintf(stderr, E_IllXparam, &Rxpara);
  return (FATAL);
}

int NothingArrive()
{
  ulong timer;
  timer = SetTimer(1);
  while (gettimer() < timer);
  if (rxlen() == 0)
    return 1;
  return 0;
}

int SendCtrl(code)
	int             code;
{
  ulong           timer;
  
  comput(code);
  timer = gettimer() + 1000L;
  while (gettimer() < timer);
}



static int RecvTerminate(LastSeqNo)
int             LastSeqNo;
{
  int             Retry;
  ulong           timer;
  int             c;
  XBLOCK          xb;
  
  FinishLog("Dnl");
  
  for (Retry = 0; Retry < MaxRetry; Retry++) {
    SendAnswerBlock(VSTAT, LastSeqNo, "0", 1);
      timer = SetTimer(RTimer);
    while ((c = comget(timer)) != TIMEOUT) {
      if (c == EOT && NothingArrive()) {
	comput(ACK);
	return (NORMAL);
      }
      if ((c == CAN) && (NothingArrive())) {
	fprintf(stderr, E_Canceled);
	return (FATAL);
      }
      if (c == SOH) {
	xb.Header = c;
	if ((c = comget(timer)) == TIMEOUT)
	  break;
	xb.Num1 = c;
	if ((c = comget(timer)) == TIMEOUT)
	  break;
	xb.Num2 = c;
	if (xb.Num1 != 0 || xb.Num2 != VFILE)
	  continue;
	if (RecvOneBlock(&xb, timer) == NORMAL) {
	  if (!isdigit(xb.Body[0]) || !isdigit(xb.Body[1])) {
	    fprintf(stderr, E_VFILE, xb.Body);
	    goto error;
	  }
	  SetFileInfo(xb.Body);
	  if (XferFileNum != FileNumber + 1) {
	    fprintf(stderr, E_VFILE, xb.Body);
	    goto error;
	  }
	  FileNumber++;
	  return (NEXT);
	}
      }
    }
  }
  
  error:
  fprintf(stderr, E_Terminate);
  SendCtrl(CAN);
  return (FATAL);
}



int qv_recieve(path)
	char	*path;
{
  XPARAM          xparam;
  int             LastSeqNo;
  int             sts;
  
  if ((sts = CheckPathname(path)) == NORMAL) {
    if ((sts = NegoSubPhase1(&xparam)) == NORMAL) {
      if ((sts = DecideXparameter(&xparam)) == NORMAL) {
	if ((sts = NegoSubPhase2(&xparam)) == NORMAL) {
	  do {
	    if ((sts = RecvStart(path)) == NORMAL) {
	      if ((sts = RecvFile(&LastSeqNo)) == NORMAL) {
		sts = RecvTerminate(LastSeqNo);
	      }
	    } else {
	      if (sts == SKIP)
		sts = NORMAL;
	    }
	  } while (sts == NEXT);
	}
      }
    }
  }
  return (sts);
}
