/*
 * $Id: qvsend.c,v 1.2 1994/09/26 18:33:41 tani Exp $
 */

#include "qv.h"
#include "qvmsg.h"

static SendFuncBlock(type, body, size)
	int             type;
	uchar          *body;
	int             size;
{
  uchar           ChkSum;
  int             remain;
  
  comput(SOH);
  comput(00);
  comput(type);
  ChkSum = SOH + type;
  for (remain = BLOCKSIZE; size > 0; size--, remain--) {
    comput(*body);
    ChkSum += *body;
    body++;
  }
  for (; remain > 0; remain--) {
    comput(00);
  }
  comput(ChkSum);
  return;
}


static int RecvResBlock(xb, timer)
	XBLOCK         *xb;
	ulong           timer;
{
  uchar           ChkSum;
  int             sum, c, i;
  
  xb->Header = STX;
  if ((c = comget(timer)) == TIMEOUT)
    goto timeout;
  xb->Num1 = c;
  if ((c = comget(timer)) == TIMEOUT)
    goto timeout;
  xb->Num2 = c;
  i = 0;
  while ((c = comget(timer)) != TIMEOUT) {
    xb->Body[i] = c;
    i++;
    if (c == CR) {
      ChkSum = xb->Body[i - 2];
      sum = xb->Header + xb->Num1 + xb->Num2;
      for (i = i - 3; i >= 0; i--) {
	sum += xb->Body[i];
      }
      if ((sum & 0x7F | 0x80) == ChkSum) {
	return (xb->Num1);
      } else {
	fprintf(stderr, E_CheckSum);
	return (ERROR);
      }
    } else {
      if (i >= BLOCKSIZE - 3) {
	fprintf(stderr, E_IllResBlock);
	return (ERROR);
      }
    }
  }
  timeout:
  fprintf(stderr, E_TimeOut);
  return (ERROR);
}


static int NegoSubPhase1()
{
  ulong           timer;
  int             c;
  int             sts;
  
  if (nopt)
    return (NORMAL);/* no waiting <NAK> option */
  timer = SetTimer(InitialWaitTime);
  for (;;) {
    c = comget(timer);
    if (c == TIMEOUT) {
      SendCtrl(CAN);
      fprintf(stderr, E_TimeOut);
      return (FATAL);
    }
    if ((c == CAN) && (NothingArrive())) {
      fprintf(stderr, E_Canceled);
      return (FATAL);
    }
    if ((c == NAK) && (NothingArrive())) {
      break;
    }
  }
  return (NORMAL);
}

static int NegoSubPhase2()
{
  static XPARAM sparam = {'1', '0', "", '1', '1', 0};
  XPARAM         *rparam;
  int             Retry;
  ulong           timer;
  int             c;
  XBLOCK          xb;
  
  sparam.WindowSIZE[0] = (WindowSizeMax / 10) + '0';
  sparam.WindowSIZE[1] = (WindowSizeMax % 10) + '0';
  SendFuncBlock(SINIT, &sparam, strlen((char *) &sparam));
  Retry = 0;
  timer = SetTimer(InitialWaitTime);
  while (Retry < MaxRetry) {
    if ((c = comget(timer)) == TIMEOUT) {
      fprintf(stderr, E_TimeOut);
      goto error;
    }
    if ((c == CAN) && (NothingArrive())) {
      fprintf(stderr, E_Canceled);
      return (FATAL);
    }
    if ((c == NAK) && (NothingArrive())) {
      SendFuncBlock(SINIT, &sparam, strlen((char *) &sparam));
      Retry++;
      timer = SetTimer(InitialWaitTime);
    }
    if (c == STX) {
      if (RecvResBlock(&xb, timer) == RINIT) {
	rparam = (XPARAM *) xb.Body;
	if ((rparam->Version != '1') ||
	    (rparam->RetryMethod != '0') ||
	    (rparam->OptionNumber != '0' &&
	     rparam->OptionNumber != '1') ||
	    (strncmp((char *)sparam.WindowSIZE, (char *)rparam->WindowSIZE, 2) < 0)) {
	  fprintf(stderr, E_IllXparam, rparam);
	  goto error;
	}
	if (rparam->OptionNumber == '0' &&
	    xb.Body[6] != CR) {
	  fprintf(stderr, E_IllXparam, rparam);
	  goto error;
	}
	if (rparam->OptionNumber == '1') {
	  if (rparam->OptionFunc1 == '1') {
	    SkipFunc = 'Y';
	  } else {
	    fprintf(stderr, E_IllXparam, rparam);
	    goto error;
	  }
	}
	if (strncmp((char *)sparam.WindowSIZE, (char *)rparam->WindowSIZE, 2) > 0) {
	  WindowSizeMax = (rparam->WindowSIZE[0] - '0') * 10 +
	    (rparam->WindowSIZE[1] - '0');
	}
	return (NORMAL);
      }
    }
  }
  error:
  fprintf(stderr, E_NEGO2Fail);
  SendCtrl(CAN);
  return (FATAL);
}

static int NegoSubPhase3(no, fname)
	int             no;
	char           *fname;
{
  static XBLOCK VFblock = {SOH, 00, VFILE};
  XBLOCK        Rxb;
  int             Retry;
  ulong           timer;
  int             i, c, skipNo;
  
  for (i = 0; i < BLOCKSIZE; i++) {
    VFblock.Body[i] = 0;
  }
  VFblock.Body[0] = (no / 10) + '0';
  VFblock.Body[1] = (no % 10) + '0';
  if (OpenSendFile(fname, &VFblock) == 0)
    return (FATAL);
  SendFuncBlock(VFILE, VFblock.Body, sizeof(VFblock.Body));
  Retry = 0;
  timer = SetTimer(STimer);
  while (Retry < MaxRetry) {
    if ((c = comget(timer)) == TIMEOUT) {
      fprintf(stderr, E_TimeOut);
      goto error;
    }
    if ((c == CAN) && (NothingArrive())) {
      fprintf(stderr, E_Canceled);
      CloseSendFile();
      return (FATAL);
    }
    if (c == STX) {
      switch (RecvResBlock(&Rxb, timer)) {
	case RINIT:
	Retry++;
	if (Retry >= MaxRetry) {
	  fprintf(stderr, E_RetryOver);
	  goto error;
	}
	SendFuncBlock(VFILE, VFblock.Body, sizeof(VFblock.Body));
	break;
	case VSKIP:
	if (SkipFunc != 'Y')
	  break;
	if (!isdigit(Rxb.Body[0]) || !isdigit(Rxb.Body[1])) {
	  Rxb.Body[2] = '\0';
	  fprintf(stderr, E_VSKIP, Rxb.Body);
	  goto error;
	}
	skipNo = (Rxb.Body[0] - '0') * 10 + (Rxb.Body[1] - '0');
	if (skipNo == no) {
	  fprintf(stderr, E_FileSkip, fname);
	  CloseSendFile();
	  return (NEXT);
	}
	if (skipNo != (no - 1)) {
	  Rxb.Body[2] = '\0';
	  fprintf(stderr, E_VSKIP, Rxb.Body);
	  goto error;
	}
	Retry++;
	if (Retry >= MaxRetry) {
	  fprintf(stderr, E_RetryOver);
	  goto error;
	}
	break;
	case VRPOS:
	if (Rxb.Body[1] == CR) {
	  StartBlockNo = 1;
	} else {
	  StartBlockNo = 0;
	  *(strchr((char *)Rxb.Body, CR) - 1) = '\0';
	  for (i = 0; Rxb.Body[i] != '\0'; i++) {
	    if (!isdigit(Rxb.Body[i])) {
	      fprintf(stderr, E_VRPOS, Rxb.Body);
	      goto error;
	    }
	    StartBlockNo = StartBlockNo * 10 + (Rxb.Body[i] - '0');
	  }
	  StartBlockNo++;
	}
	if (StartBlockNo > LastBlockNo) {
	  fprintf(stderr, E_VRPOS, Rxb.Body);
	  goto error;
	}
	return (NORMAL);
      }
    }
  }
  error:
  fprintf(stderr, E_NEGO3Fail);
  CloseSendFile();
  SendCtrl(CAN);
  return (FATAL);
}


static int SendOneBlock(SeqNo)
	int             SeqNo;
{
  unsigned char   block_buf[BLOCKSIZE];
  int             size;
  unsigned char   Num, ChkSum;
  
  if ((size = ReadOneBlock(StartBlockNo + SeqNo - 1, block_buf)) == 0) {
    return (FATAL);
  }
  if (size < BLOCKSIZE) {
    for (; size < BLOCKSIZE; size++) {
      block_buf[size] = PADCHAR;
    }
  }
  comput(SOH);
  Num = SeqNo;
  comput(Num);
  comput(~Num);
  ChkSum = 0;
  comwrite(block_buf, BLOCKSIZE);
  for (size = 0; size < BLOCKSIZE; size++) {
    ChkSum += block_buf[size];
  }
  comput(ChkSum);
  XferTotalBlock++;
  return (NORMAL);
}

static int SendFile()
{
  int             j = 0;
  int             vS = 0;
  int             k = 0;
  int             m;
  ulong           timer;
  int             Retry = 0;
  int             c, Check;
  int             ENQ_flag = 0;
  XBLOCK          Rxb;
  static struct {
    uchar           Num;
    uchar           Func;
  }               venq = {
    0, '0'
    };
  
  XferStart = gettimer();
  XferBlock = 0;
  XferTotalBlock = 0;
  XferRetry = 0;
  for (;;) {
    while (StartBlockNo + j <= LastBlockNo && k < WindowSizeMax) {
      k++;
      j++;
      if (SendOneBlock(j) != NORMAL)
	return (FATAL);
      if (rxlen() > 0)
	break;
    }
    if (StartBlockNo + j - 1 == LastBlockNo && ENQ_flag == 0) {
      venq.Num = j;
      SendFuncBlock(VENQ, &venq, 2);
      ENQ_flag = 1;
    }
    timer = SetTimer(STimer);
    Check = 1;
    while (Check) {
      if ((c = comget(timer)) == TIMEOUT) {
	SendCtrl(CAN);
	fprintf(stderr, E_TimeOut);
	return (FATAL);
      }
      if ((c == CAN) && NothingArrive()) {
	SendCtrl(EOT);
	fprintf(stderr, E_Canceled);
	return (FATAL);
      }
      if (c == STX) {
	c = RecvResBlock(&Rxb, timer);
	m = (Rxb.Num2 & 0x7F) + (vS & 0xFF80);
	if (m < vS)
	  m += 0x80;
	switch (c) {
	  case VACK:
	  if (m > vS && m <= j) {
	    XferBlock = m;
	    k = j - m;
	    vS = m;
	    Retry = 0;
	    Check = 0;
	    DisplayStatus();
	  }
	  break;
	  case VNAK:
	  if (Retry == 0) {
	    if (m > vS && m <= j + 1) {
	      Retry++;
	      j = vS = m - 1;
	      k = 0;
	      ENQ_flag = 0;
	      Check = 0;
	    }
	  } else {
	    if (m >= vS + 1 && m <= j + 1) {
	      if (m == vS + 1) {
		Retry++;
	      } else {
		Retry = 1;
	      }
	      j = vS = m - 1;
	      k = 0;
	      ENQ_flag = 0;
	      Check = 0;
	    }
	  }
	  if (Retry > SCount) {
	    SendCtrl(CAN);
	    fprintf(stderr, E_RetryOver);
	    return (FATAL);
	  }
	  if (Check == 0) {
	    XferRetry++;
	  }
	  break;
	  case VSTAT:
	  if (m != j)
	    break;
	  if (ENQ_flag && Rxb.Body[0] == '0') {
	    XferBlock = m;
	    XferFinish = gettimer();
	    FinishLog("Upl");
	    return (NORMAL);
	  } else {
	    fprintf(stderr, E_IllProtocol, "VSTAT");
	    return (FATAL);
	  }
	  case VRPOS:
	  if (vS == 0) {
	    j = k = 0;
	    Retry++;
	    Check = 0;
	    XferRetry++;
	  }
	  if (Retry > SCount) {
	    SendCtrl(CAN);
	    fprintf(stderr, E_RetryOver);
	    return (FATAL);
	  }
	  break;
	}
      }
    }
  }
}

static int SendTerminate()
{
  ulong           timer;
  int             c, Retry;
  
  Retry = 0;
  SendCtrl(EOT);
  timer = SetTimer(STimer);
  while ((c = comget(timer)) != TIMEOUT) {
    if (c == ACK && NothingArrive()) {
      return (NORMAL);
    }
    if (c == CAN && NothingArrive()) {
      fprintf(stderr, E_Canceled);
      return (FATAL);
    }
    Retry++;
    if (Retry > SCount) {
      SendCtrl(CAN);
      return (FATAL);
    }
    SendCtrl(EOT);
    timer = SetTimer(STimer);
  }
  SendCtrl(CAN);
  return (FATAL);
}


int qv_send(n, fname)
	int             n;
	char           *fname[];
{
  int             i, sts;
  
  if ((n = CheckSendFiles(n, fname)) == 0) {
    SendCtrl(CAN);
    return (ERROR);
  }
  if ((sts = NegoSubPhase1()) == NORMAL) {
    if ((sts = NegoSubPhase2()) == NORMAL) {
      for (i = 0; i < n; i++) {
	if ((sts = NegoSubPhase3(i + 1, fname[i])) == FATAL)
	  goto fin;
	if (sts == NORMAL) {
	  sts = SendFile();
	  CloseSendFile();
	  if (sts != NORMAL)
	    goto fin;
	}
      }
      sts = SendTerminate();
    }
  }
  fin:
  return (sts);
}
