/*
 * $Id: telnetx.c,v 1.1 1995/11/28 01:07:48 kuroda Exp kuroda $
 * Copyright (c) 1993  The University of Electro-Communications
 * All rights reserved.
 */

#ifdef TELNETX

#include <stdio.h>

#ifndef luna
#include <unistd.h>
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/signal.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>

#ifdef USE_TERMIO
#include <termios.h>
#endif

#ifdef	STREAMSPTY
#include <sac.h>
#include <sys/stropts.h>
#endif

#ifndef luna
#include <setjmp.h>
#endif

#include "ring.h"
#include "externs.h"
#include "defines.h"

/* it is as same as ring.c(ring_empty) */
#define	ring_empty(d) (((d)->consume == (d)->supply) && \
				((d)->consumetime >= (d)->supplytime))

static int	stdin_back;
static int	stdout_back;
#if defined(sony) || defined(luna)
typedef int pid_t;
#endif
static pid_t	child_pid;
static int	pty;
static int	intr_count = 0;	/* key interrupt count */

static void (*intr)();
static void (*child)();
#ifdef hpux
static void (*sig_pipe)();
#endif

/*
 * zm_getpty()
 */

#ifdef STREAMSPTY

int zm_getpty(pty_line)
char *pty_line;
{
  int p;

  (void) sprintf(pty_line, "/dev/ptmx");
  p = open(pty_line, O_RDWR);
  if (p > 0) {
    grantpt(p);
    unlockpt(p);
    strcpy(pty_line, ptsname(p));
    return(p);
  }
  return(-1);
}

#else

int zm_getpty(pty_line)
char *pty_line;
{
  register int p;
  register char c, *p1, *p2;
  register int i;
  
  (void) sprintf(pty_line, "/dev/ptyXX");
  p1 = &pty_line[8];
  p2 = &pty_line[9];
  
  for (c = 'p'; c <= 's'; c++) {
    struct stat stb;
    
    *p1 = c;
    *p2 = '0';
    if (stat(pty_line, &stb) < 0)
      break;
    for (i = 0; i < 16; i++) {
      *p2 = "0123456789abcdef"[i];
      p = open(pty_line, O_RDWR);
      if (p > 0) {
	pty_line[5] = 't';
	return(p);
      }
    }
  }
  return(-1);
}

#endif	/* STREAMSPTY  */

/* Buffer state transition routines */

ring_clear(ring)
Ring *ring;
{
    ring->supply = ring->consume = ring->bottom;
    ring->consumetime = ring->supplytime;
    memset((char *)ring->bottom, 0, ring->size);
#if	defined(ENCRYPTION)
    ring->clearto = 0;
#endif
    return 1;
}

void zm_child_died()
{
  int tty;

  (void)wait((int *)0);
  signal(SIGCHLD, child);
  signal(SIGINT, intr);
#ifdef hpux
  signal(SIGPIPE, sig_pipe);
#endif
  intr_count = 0;

  if (!ring_empty(&netoring)) {
    ring_clear(&netoring);
  }
  if (!ring_empty(&netiring)) {
    ring_clear(&netiring);
  }
  if (!ring_empty(&ttyoring)) {
    ring_clear(&ttyoring);
  }
  if (!ring_empty(&ttyiring)) {
    ring_clear(&ttyiring);
  }
  dup2(stdin_back,  0); 
  dup2(stdout_back, 1);
  close(pty);
  close(stdin_back);
  close(stdout_back);
  in_zmodem = 0;
#if defined(FNET)
  tel_enter_binary(eight);
#endif
  command(0, "\n", 1);
}

void zm_kill_child()
{
  if (intr_count++ > 3) {
    printf("kill child !!\n");
    kill(child_pid, SIGKILL);
  } else {
    kill(child_pid, SIGTERM);
  }
}

zm_rings(pid, m_pty)
int pid;
int m_pty;
{
  child_pid = pid;
  pty = m_pty;
  intr_count = 0;

  intr = signal(SIGINT, zm_kill_child);
  child = signal(SIGCHLD, zm_child_died);
#ifdef hpux
  sig_pipe = signal(SIGPIPE, zm_child_died);
#endif
  stdin_back = dup(0);
  stdout_back = dup(1);
  dup2(m_pty, 0);
  dup2(m_pty, 1);
  in_zmodem = 1;
}

#endif /* TELNETX */

/* ---- end of zmodem.c ---- */
